""" Module for FusionCatcher parser """
from __future__ import annotations
import re
from typing import Iterable, List, Tuple
import itertools
from moPepGen.SeqFeature import FeatureLocation
from moPepGen import seqvar, gtf, dna, err


def parse(path:str) -> Iterable[FusionCatcherRecord]:
    """ Parse the FusionCatcher's output and returns an iterator.

    Args:
        path (str): Path to the FusionCatcher's output file.
    """
    with open(path, 'r') as handle:
        # first line is header
        line = next(handle, None)
        line = next(handle, None)
        while line:
            if line.startswith('#'):
                line = next(handle, None)
                continue
            fields = line.rstrip().split('\t')
            yield FusionCatcherRecord(
                five_end_gene_symbol=fields[0],
                three_end_gene_symbol=fields[1],
                fusion_descriptions=fields[2].split(','),
                counts_of_common_mapping_reads=int(fields[3]),
                spanning_pairs=int(fields[4]),
                spanning_unique_reads=int(fields[5]),
                longest_anchor_found=int(fields[6]),
                fusion_finding_method=fields[7].split(';'),
                five_end_breakpoint=fields[8],
                three_end_breakpoint=fields[9],
                five_end_gene_id=fields[10],
                three_end_gene_id=fields[11],
                five_end_exon_id=fields[12],
                three_end_exon_id=fields[13],
                fusion_sequence=tuple(fields[14].split('*')),
                predicted_effect=fields[15]
            )
            line = next(handle, None)


class FusionCatcherRecord():
    """ Defines a record of FusionCatcher's output. """
    def __init__(self, five_end_gene_symbol:str, three_end_gene_symbol:str,
            fusion_descriptions:List[str], counts_of_common_mapping_reads:int,
            spanning_pairs:int, spanning_unique_reads:int,
            longest_anchor_found:int, fusion_finding_method:List[str],
            five_end_breakpoint:str, three_end_breakpoint:str,
            five_end_gene_id:str, three_end_gene_id:str, five_end_exon_id:str,
            three_end_exon_id:str, fusion_sequence:Tuple[str, str],
            predicted_effect:str
            ):
        """"""
        self.five_end_gene_symbol = five_end_gene_symbol
        self.three_end_gene_symbol = three_end_gene_symbol
        self.fusion_descriptions = fusion_descriptions
        self.counts_of_common_mapping_reads = counts_of_common_mapping_reads
        self.spanning_pairs = spanning_pairs
        self.spanning_unique_reads = spanning_unique_reads
        self.longest_anchor_found = longest_anchor_found
        self.fusion_finding_method = fusion_finding_method
        self.five_end_breakpoint = five_end_breakpoint
        self.three_end_breakpoint = three_end_breakpoint
        self.five_end_gene_id = five_end_gene_id
        self.three_end_gene_id = three_end_gene_id
        self.five_end_exon_id = five_end_exon_id
        self.three_end_exon_id = three_end_exon_id
        self.fusion_sequence = fusion_sequence
        self.predicted_effect = predicted_effect

    @property
    def left_breakpoint_position(self) -> int:
        """ Get left breakpoint position """
        return int(self.five_end_breakpoint.split(':')[1])

    @property
    def right_breakpoint_position(self) -> int:
        """ Get right breakpoint position """
        return int(self.three_end_breakpoint.split(':')[1])

    def get_donor_transcripts(self, anno:gtf.GenomicAnnotation, gene_id:str
            ) -> List[gtf.TranscriptAnnotationModel]:
        """ Get all possible donor transcripts """
        pos = self.left_breakpoint_position - 1
        return anno.get_transcripts_with_position(gene_id, pos)

    def get_accepter_transcripts(self, anno:gtf.GenomicAnnotation, gene_id:str
            ) -> List[gtf.TranscriptAnnotationModel]:
        """ Get all possible accepter transcripts """
        pos = self.right_breakpoint_position - 1
        return anno.get_transcripts_with_position(gene_id, pos)

    def convert_to_variant_records(self, anno:gtf.GenomicAnnotation,
            genome:dna.DNASeqDict) -> List[seqvar.VariantRecord]:
        """ COnvert a FusionCatcher's record to VariantRecord.

        Args:
            anno (Dict[str, Dict[str, gtf.TranscriptAnnotationModel]]): The
                genome annotation. The first level key is the gene ID, and the
                second level key is the transcript ID.
            genome (dna.DNASeqDict): The genome sequence records.
        Returns:
            List of VariantRecord
        """
        pattern = re.compile(r'\.[0-9]+$')
        if pattern.search(self.five_end_gene_id):
            donor_gene_id = self.five_end_gene_id
            accepter_gene_id = self.three_end_gene_id
            try:
                donor_gene_model = anno.genes[self.five_end_gene_id]
            except KeyError as error:
                raise err.GeneNotFoundError(self.five_end_gene_id) from error
            try:
                accepter_gene_model = anno.genes[self.three_end_gene_id]
            except KeyError as error:
                raise err.GeneNotFoundError(self.three_end_gene_id) from error
        else:
            donor_gene_model = anno.get_gene_model_from_unversioned_id(
                self.five_end_gene_id)
            accepter_gene_model = anno.get_gene_model_from_unversioned_id(
                self.three_end_gene_id)
            donor_gene_id = donor_gene_model.gene_id
            accepter_gene_id = accepter_gene_model.gene_id

        # in case the ensembl ID does not match those in annotation
        # and not the same gene is referered to
        donor_gene_symbol = donor_gene_model.gene_name
        accepter_gene_symbol = accepter_gene_model.gene_name

        donor_chrom = donor_gene_model.chrom
        accepter_chrom = accepter_gene_model.chrom

        # fusion catcher uses 1-based coordinates
        # left breakpoint is the first nucleotide in the fusion transcript
        # after the breakpoint
        left_breakpoint_genomic = self.left_breakpoint_position
        right_breakpoint_genomic = self.right_breakpoint_position
        donor_genome_position = \
            f'{donor_chrom}:{left_breakpoint_genomic}:{left_breakpoint_genomic}'
        accepter_genome_position = \
            f'{accepter_chrom}:{right_breakpoint_genomic}:{right_breakpoint_genomic}'
        left_breakpoint_genetic = anno.coordinate_genomic_to_gene(
            index=left_breakpoint_genomic - 1, gene=donor_gene_id
        ) + 1
        right_breakpoint_genetic = anno.coordinate_genomic_to_gene(
            index=right_breakpoint_genomic - 1, gene=accepter_gene_id
        )

        donor_transcripts = self.get_donor_transcripts(anno, donor_gene_id)
        accepter_transcripts = self.get_accepter_transcripts(anno, accepter_gene_id)

        records = []

        if donor_gene_model.strand == 1:
            ref_seq = genome[donor_chrom].seq[left_breakpoint_genomic]
        else:
            ref_seq = genome[donor_chrom]\
                .seq[left_breakpoint_genomic:left_breakpoint_genomic + 1]\
                .reverse_complement()
            ref_seq = str(ref_seq)

        perms = itertools.product(donor_transcripts, accepter_transcripts)
        for donor_tx, accepter_tx in perms:
            donor_tx_id = donor_tx.transcript.transcript_id
            accepter_tx_id = accepter_tx.transcript.transcript_id

            fusion_id = f'FUSION-{donor_tx_id}:{left_breakpoint_genetic}'+\
                f'-{accepter_tx_id}:{right_breakpoint_genetic}'

            location = FeatureLocation(
                seqname=donor_gene_id,
                start=left_breakpoint_genetic,
                end=left_breakpoint_genetic + 1
            )
            attrs = {
                'TRANSCRIPT_ID': donor_tx_id,
                'GENE_SYMBOL': donor_gene_symbol,
                'GENOMIC_POSITION': donor_genome_position,
                'ACCEPTER_GENE_ID': accepter_gene_id,
                'ACCEPTER_TRANSCRIPT_ID': accepter_tx_id,
                'ACCEPTER_SYMBOL': accepter_gene_symbol,
                'ACCEPTER_POSITION': right_breakpoint_genetic,
                'ACCEPTER_GENOMIC_POSITION': accepter_genome_position
            }

            record = seqvar.VariantRecord(
                location=location,
                ref=ref_seq,
                alt='<FUSION>',
                _type='Fusion',
                _id=fusion_id,
                attrs=attrs
            )
            records.append(record)
        return records

    def set_chrom_with_chr(self, with_chr=True) -> None:
        """ Depending on if chr is required in chrom, fix breakpoints"""
        if with_chr:
            if 'chr' not in self.five_end_breakpoint.split(':')[0]:
                self.five_end_breakpoint = 'chr' + self.five_end_breakpoint
                self.three_end_breakpoint = 'chr' + self.three_end_breakpoint
        if not with_chr:
            if 'chr' in self.five_end_breakpoint.split(':')[0]:
                feb = self.five_end_breakpoint
                feb = feb[feb.startswith('chr') and len('chr'):]
                teb = self.three_end_breakpoint
                teb = teb[teb.startswith('chr') and len('chr'):]
                self.five_end_breakpoint = feb
                self.three_end_breakpoint = teb



""" Version """
from __future__ import annotations
import sys
from typing import Tuple
import Bio
from moPepGen import __version__


MINIMAL_VERSION = '1.3.0'

class MetaVersion():
    """ Versions """
    def __init__(self, python:str=None, biopython:str=None, mopepgen:str=None):
        """ constructor """
        self.python = python or '.'.join([str(x) for x in sys.version_info[:3]])
        self.biopython = biopython or Bio.__version__
        self.mopepgen = mopepgen or __version__

    def __eq__(self, other:MetaVersion):
        """ equal to """
        return self.python == other.python and \
            self.biopython == other.biopython and \
            self.mopepgen == other.mopepgen

    @staticmethod
    def get_semver(version:str) -> Tuple[int, int, int]:
        """ Get major, minor and patch number from semver """
        return tuple(int(x) for x in version.split('-')[0].split('.'))

    def is_valid_mpg_version(self, version:str) -> bool:
        """ Check if the given moPepGen version is valid """
        that = self.get_semver(version)
        minimal = self.get_semver(MINIMAL_VERSION)
        return that >= minimal

    def is_valid(self, version:MetaVersion) -> bool:
        """ Check if the given version is valid """
        return self.python == version.python and \
            self.biopython == version.biopython and \
            self.is_valid_mpg_version(version.mopepgen)

    def __ne__(self, other:MetaVersion):
        """ not equal to """
        return not self == other

    def __repr__(self) -> str:
        """ str representation """
        return f"python={self.python}, biopython={self.biopython}, moPepGen={self.mopepgen}"

    def jsonfy(self):
        """ jsonfy """
        return {
            'python': self.python,
            'biopython': self.biopython,
            'mopepgen': self.mopepgen
        }



""" Module for the squence variant graph """
from moPepGen.svgraph.TVGNode import TVGNode
from moPepGen.svgraph.TVGEdge import TVGEdge
from moPepGen.svgraph.ThreeFrameTVG import ThreeFrameTVG
from moPepGen.svgraph.TVGCursor import TVGCursor
from moPepGen.svgraph.PeptideVariantGraph import PeptideVariantGraph
from moPepGen.svgraph.PVGNode import PVGNode
from moPepGen.svgraph.ThreeFrameCVG import ThreeFrameCVG
from moPepGen.svgraph.VariantPeptideTable import VariantPeptideTable



""" This module defines the class for a single VEP record
"""
from __future__ import annotations
from typing import List, Tuple, Iterable, IO
from Bio.Seq import Seq
from moPepGen.SeqFeature import FeatureLocation
from moPepGen.err import TranscriptionStopSiteMutationError, \
    TranscriptionStartSiteMutationError
from moPepGen import seqvar, dna, gtf


def parse(handle:IO) -> Iterable[VEPRecord]:
    """ Parse a VEP output text file and return as an iterator.

    Args:
        path (str): Path to the REDItools output table.

    Return:
        A iterable of VEPRecord.
    """
    for line in handle:
        if line.startswith('#'):
            continue
        line = line.rstrip()
        fields = line.split('\t')

        consequences = fields[6].split(',')

        amino_acids = tuple(aa for aa in fields[10].split('/'))
        if len(amino_acids) == 1:
            amino_acids = (amino_acids[0], '')

        codons = tuple(codon for codon in fields[11].split('/'))
        if len(codons) == 1:
            codons = (codons[0], '')

        extra = {}

        yield VEPRecord(
            uploaded_variation=fields[0],
            location=fields[1],
            allele=fields[2],
            gene=fields[3],
            feature=fields[4],
            feature_type=fields[5],
            consequences=consequences,
            cdna_position=fields[7],
            cds_position=fields[8],
            protein_position=fields[9],
            amino_acids=amino_acids,
            codons=codons,
            existing_variation='' if fields[12] == '-' else fields[12],
            extra=extra
        )

class VEPRecord():
    """ A VEPRecord object holds the an entry from the VEP output. The VEP
    output is defined at https://uswest.ensembl.org/info/docs/tools/vep/
    vep_formats.html#output

    Attributes:
        uploaded_variation (str): as chromosome_start_alleles
        location (str): in standard coordinate format (chr:start or
            chr:start-end)
        allele (str): the variant allele used to calculate the consequence
        gene (str): Ensembl stable ID of affected gene
        feature (str): Ensembl stable ID of feature
        feature_type (str): type of feature. Currently one of Transcript,
            RegulatoryFeature, MotifFeature.
        consequence (List[str]): consequence type of this variant.
            See: https://uswest.ensembl.org/info/genome/variation/prediction/
            predicted_data.html#consequences
        cdna_position (str): relative position of base pair in cDNA sequence
        cds_position (str): relative position of base pair in coding sequence
        protein_position (str): relative position of amino acid in protein
        amino_acids (Tuple[str]): only given if the variant affects the
            protein-coding sequence
        codons (Tuple[str]) the alternative codons with the variant base in
            upper case
        existing_variation (str) known identifier of existing variant
        extra (dict): this column contains extra information.
    """
    def __init__(
            self, uploaded_variation: str, location: str, allele: str,
            gene: str, feature: str, feature_type:str,
            consequences: List[str], cdna_position: str, cds_position: str,
            protein_position: str, amino_acids: Tuple[str, str],
            codons: Tuple[str, str], existing_variation: str, extra: dict):
        """ Construct a VEPRecord object. """
        self.uploaded_variation = uploaded_variation
        self.location = location
        self.allele = allele
        self.gene = gene
        self.feature = feature
        self.feature_type=feature_type
        self.consequences = consequences
        self.cdna_position = cdna_position
        self.cds_position = cds_position
        self.protein_position = protein_position
        self.amino_acids = amino_acids
        self.codons = codons
        self.existing_variation = existing_variation
        self.extra = extra

    def __repr__(self)->str:
        """Return representation of the VEP record."""
        consequences = '|'.join(self.consequences)
        return f"< {self.feature}, {consequences}, {self.location}, {self.allele} >"

    def convert_to_variant_record(self, anno:gtf.GenomicAnnotation,
            genome:dna.DNASeqDict) -> seqvar.VariantRecord:
        """ Convert a VepRecord to a generic VariantRecord object.

        Args:
            seq (dna.DNASeqRecord): The DNA sequence of the transcript.
        """
        _, alt_position = self.location.split(':')
        if alt_position.find('-') > -1:
            alt_start_genomic, alt_end_genomic = \
                [int(x) for x in alt_position.split('-')]
        else:
            alt_start_genomic = int(alt_position)
            alt_end_genomic = alt_start_genomic
        alt_start_genomic -= 1

        gene_model = anno.genes[self.gene]
        chrom_seqname = gene_model.chrom
        strand = gene_model.strand
        seq = gene_model.get_gene_sequence(genome[chrom_seqname])

        alt_start = anno.coordinate_genomic_to_gene(alt_start_genomic, self.gene)
        alt_end = anno.coordinate_genomic_to_gene(alt_end_genomic - 1, self.gene)

        tx_id = self.feature
        tx_model = anno.transcripts[tx_id]

        if strand == 1:
            tx_first = tx_model.transcript.location.start
            tx_last = tx_model.transcript.location.end - 1
            tx_start_genetic = anno.coordinate_genomic_to_gene(tx_first, self.gene)
            tx_end_genetic = anno.coordinate_genomic_to_gene(tx_last, self.gene) + 1
        else:
            tx_first = tx_model.transcript.location.end - 1
            tx_last = tx_model.transcript.location.start
            tx_start_genetic = anno.coordinate_genomic_to_gene(tx_first, self.gene)
            tx_end_genetic = anno.coordinate_genomic_to_gene(tx_last, self.gene) + 1

        if strand == -1:
            alt_start, alt_end = alt_end, alt_start
        alt_end += 1

        if alt_start < tx_start_genetic or (alt_start == tx_start_genetic \
                and not tx_model.is_cds_start_nf()):
            raise TranscriptionStartSiteMutationError(tx_id)

        if alt_end > tx_end_genetic:
            raise TranscriptionStopSiteMutationError(tx_id)

        if self.allele == '-':
            if alt_start == tx_start_genetic:
                alt_end += 1
                ref = str(seq.seq[alt_start:alt_end])
                alt = str(seq.seq[alt_end-1:alt_end])
            else:
                alt_start -= 1
                ref = str(seq.seq[alt_start:alt_end])
                alt = str(seq.seq[alt_start])
        else:
            allele = self.allele
            if strand == -1:
                allele = str(Seq(allele).reverse_complement())
            if alt_end - alt_start == 1:
                if len(allele) > 1: # insertion
                    # Sometimes insertions are reported by VEP in the end-inclusion
                    # way (e.g., C -> TACC), which needs to be converted into
                    # start-inclusion (A -> ATAC) for variants on + strand genes.
                    ref = str(seq.seq[alt_start])
                    if ref == allele[-1]:
                        alt_start -= 1
                        alt_end = alt_start + 1
                        ref = str(seq.seq[alt_start])
                        alt = ref + allele[:-1]
                    elif ref == allele[0]:
                        ref = str(ref)
                        alt = allele
                    else:
                        raise ValueError(f"Don't know how to process this variant: {self}")
                else: # SNV
                    ref = str(seq.seq[alt_start])
                    alt = allele
            elif alt_end - alt_start == 2:  # insertion
                alt_end -= 1
                ref = str(seq.seq[alt_start])
                alt = ref + allele
            else:
                ref = str(seq.seq[alt_start:alt_end])
                alt = allele

        if len(ref) == len(alt) == 1:
            _type = 'SNV'
        elif len(ref) == 1 or len(ref) == 1:
            _type = 'INDEL'
        else:
            _type = 'MNV'
        _id = f'{_type}-{alt_start + 1}-{ref}-{alt}'

        attrs = {
            'TRANSCRIPT_ID': self.feature,
            'GENOMIC_POSITION': self.location,
            'GENE_SYMBOL': gene_model.gene_name
        }

        try:
            return seqvar.VariantRecord(
                location=FeatureLocation(
                    seqname=self.gene,
                    start=alt_start,
                    end=alt_end
                ),
                ref=ref,
                alt=alt,
                _type=_type,
                _id=_id,
                attrs=attrs
            )
        except ValueError as e:
            raise ValueError(e.args[0] + f' [{self.feature}]') from e



""" Module for TVGNode class """
from __future__ import annotations
from typing import List, Set, Tuple, Dict, Deque, TYPE_CHECKING
import copy
from collections import deque
import math
from Bio.Seq import Seq
from moPepGen.dna import DNASeqRecordWithCoordinates
from moPepGen import seqvar, aa
from moPepGen.SeqFeature import FeatureLocation, MatchedLocation
from .PVGNode import PVGNode
from .TVGEdge import TVGEdge

if TYPE_CHECKING:
    from moPepGen.svgraph import SubgraphTree

class TVGNode():
    """ Defines the nodes in the TranscriptVariantGraph

    Attributes:
        in_edges (Set[TVGEdge]): The inbonding edges.
        out_edges (Set[TVGEdge]): The outbonding edges.
        seq (DNASeqRecord): The sequence.
        variant (VariantRecord | None): The variant record or None for
            reference.
        branch (bool)
    """
    def __init__(self, seq:DNASeqRecordWithCoordinates,
            variants:List[seqvar.VariantRecordWithCoordinate]=None,
            branch:bool=False, orf:List[int]=None, reading_frame_index:int=None,
            subgraph_id:str=None, global_variant:seqvar.VariantRecord=None,
            level:int=0, was_bridge:bool=False):
        """ Constructor for TVGNode.

        Args:
            seq (DNASeqRecord): The sequence.
            variant (VariantRecord | None): The variant record or None for
                reference.
        """
        self.seq = seq
        self.in_edges:Set[TVGEdge] = set()
        self.out_edges:Set[TVGEdge] = set()
        self.variants:List[seqvar.VariantRecordWithCoordinate] = variants if variants else []
        self.branch = branch
        self.orf = orf or [None, None]
        self.reading_frame_index = reading_frame_index
        self.subgraph_id = subgraph_id
        self.global_variant = global_variant
        self.level = level
        self.was_bridge = was_bridge

    def create_node(self, seq:DNASeqRecordWithCoordinates,
            variants:List[seqvar.VariantRecordWithCoordinate]=None,
            branch:bool=False, orf:List[int]=None, reading_frame_index:int=None,
            subgraph_id:str=None, global_variant:seqvar.VariantRecord=None,
            level:int=None, was_bridge:bool=False):
        """ Constructor for TVGNode.

        Args:
            seq (DNASeqRecord): The sequence.
            variant (VariantRecord | None): The variant record or None for
                reference.
        """
        return self.__class__(
            seq=seq,
            variants=variants,
            branch=branch,
            orf=orf,
            reading_frame_index=reading_frame_index,
            subgraph_id=subgraph_id,
            global_variant=global_variant or self.global_variant,
            level=level or self.level,
            was_bridge=was_bridge
        )

    def __hash__(self):
        """ hash """
        if self.seq:
            seq = self.seq.seq
            locations = [(it.ref.start, it.ref.end) for it in \
                self.seq.locations]
        else:
            seq = None
            locations = []
        return hash((str(seq), *locations))

    def __getitem__(self, index) -> TVGNode:
        """ get item """
        start, stop, _ = index.indices(len(self.seq))
        location = FeatureLocation(start=start, end=stop)
        seq = self.seq.__getitem__(index)
        variants = []

        for variant in self.variants:
            if variant.location.overlaps(location):
                new_start = max(variant.location.start, location.start)
                new_end = min(variant.location.end, location.end)
                new_start_offset = variant.location.start_offset \
                    if new_start == variant.location.start else 0
                new_end_offset = variant.location.end_offset \
                    if new_end == variant.location.end else 0
                new_loc = FeatureLocation(
                    start=max(variant.location.start, location.start),
                    end=min(variant.location.end, location.end),
                    seqname=variant.location.seqname,
                    reading_frame_index=variant.location.reading_frame_index,
                    start_offset=new_start_offset,
                    end_offset=new_end_offset
                )
                new_variant = seqvar.VariantRecordWithCoordinate(
                    variant=variant.variant, location=new_loc
                )
                new_variant = new_variant.shift(-start)
                variants.append(new_variant)
        return self.create_node(
            seq=seq,
            variants=variants,
            orf=self.orf,
            branch=self.branch,
            reading_frame_index=self.reading_frame_index,
            subgraph_id=self.subgraph_id,
            level=self.level
        )

    def is_bridge(self) -> None:
        """ Check if this is a bridge node to another reading frame """
        for edge in self.out_edges:
            node = edge.out_node
            if node.reading_frame_index != self.reading_frame_index:
                return True
        if self.was_bridge:
            return True
        return False

    def get_edge_to(self, other:TVGNode) -> TVGEdge:
        """ Find the edge from this to the other node """
        for edge in self.out_edges:
            if edge.out_node is other:
                return edge
        raise ValueError('TVGEdge not found')

    def get_edge_from(self, other:TVGNode) -> TVGEdge:
        """ Find the edge from the other to this node. """
        for edge in self.in_edges:
            if edge.in_node is other:
                return edge
        raise ValueError('TVGEdge not found')

    def get_out_nodes(self) -> List[TVGNode]:
        """ get outbonding nodes as a list """
        return [x.out_node for x in self.out_edges]

    def get_in_nodes(self) -> List[TVGNode]:
        """ get incoming nodes as a list """
        return [x.in_node for x in self.in_edges]

    def _get_nth_rf_index(self, i:int) -> int:
        """ """
        if (i > 0 or i < -1) and not self.has_multiple_segments():
            raise ValueError('Node does not have multiple segments')

        locations = [loc.query for loc in self.seq.locations]
        locations += [v.location for v in self.variants
            if self.global_variant is None or v.variant != self.global_variant]

        locations.sort()

        return locations[i].reading_frame_index

    def get_first_rf_index(self) -> int:
        """ Get the first fragment's reading frame index """
        return self._get_nth_rf_index(0)

    def get_second_rf_index(self) -> int:
        """ Get the second fragment's reading frame index """
        return self._get_nth_rf_index(1)

    def get_last_rf_index(self) -> int:
        """ Get the last fragment's reading frame index """
        return self._get_nth_rf_index(-1)

    def _get_nth_subgraph_id(self, i) -> str:
        """ Get the nth fragment's subgraph ID """
        if (i > 0 or i < -1) and not self.has_multiple_segments():
            raise ValueError('Node does not have multiple segments')

        locations = [(loc.query, loc.ref.seqname) for loc in self.seq.locations]
        locations += [(v.location, v.location.seqname) for v in self.variants
            if self.global_variant is None or v.variant != self.global_variant]

        locations = sorted(locations, key=lambda x: x[0])

        return locations[i][1]

    def get_max_subgraph_id(self, subgraphs:SubgraphTree, last:bool=False) -> str:
        """ Get the max subgraph ID """
        max_subgraph_id = None
        max_level = -1
        if last:
            comp = lambda x,y: x >= y
        else:
            comp = lambda x,y: x > y
        for loc in self.seq.locations:
            subgraph_id = loc.ref.seqname
            level = subgraphs[subgraph_id].level
            if comp(level, max_level):
                max_level = level
                max_subgraph_id = subgraph_id
        for v in self.variants:
            subgraph_id = v.location.seqname
            level = subgraphs[subgraph_id].level
            if comp(level, max_level):
                max_subgraph_id = subgraph_id

        if not max_subgraph_id:
            max_subgraph_id = self.subgraph_id

        return max_subgraph_id

    def get_min_subgraph_id(self, subgraphs:SubgraphTree) -> str:
        """ Get the min subgraph ID """
        min_subgraph_id = None
        min_level = float('inf')
        for loc in self.seq.locations:
            subgraph_id = loc.ref.seqname
            level = subgraphs[subgraph_id].level
            if level < min_level:
                min_subgraph_id = subgraph_id
        for v in self.variants:
            subgraph_id = v.location.seqname
            level = subgraphs[subgraph_id].level
            if level < min_level:
                min_subgraph_id = subgraph_id

        if not min_subgraph_id:
            min_subgraph_id = self.subgraph_id

        return min_subgraph_id

    def get_first_subgraph_id(self) -> str:
        """ Get the first fragment's subgraph ID """
        return self._get_nth_subgraph_id(0)

    def get_last_subgraph_id(self) -> str:
        """ Get the last fragment's subgraph ID """
        return self._get_nth_subgraph_id(-1)

    def has_multiple_segments(self) -> bool:
        """ Whether the node has multiple segments, which is when the node
        is merged from several individual nodes. """
        n = len(self.seq.locations)
        n += len([v for v in self.variants if self.global_variant is not None
            and v.variant != self.global_variant])
        return n > 1

    def is_inbond_of(self, node:TVGEdge) -> bool:
        """ Checks if this node is a inbound node of the other node. """
        for edge in self.out_edges:
            if node is edge.out_node:
                return True
        return False

    def is_orphan(self) -> bool:
        """ Checks if the node is an orphan, that doesn't have any inbonding
        or outbonding edge. """
        return (not self.in_edges) and (not self.out_edges)

    def is_reference(self) -> bool:
        """ check if it is reference (no variants) """
        if len({x.variant for x in self.variants}) == 1 \
                and any(x.variant.is_real_fusion for x in self.variants):
            return True
        if self.global_variant is None:
            return not self.variants
        return not any(v.variant is not self.global_variant and
            v.variant != self.global_variant for v in self.variants)

    def is_stop_node(self) -> bool:
        """ check if it is a stop node """
        return self.seq.seq == '' and not self.out_edges

    def has_in_bridge(self) -> bool:
        """ check if it has any in node from different reading frame """
        return any(e.in_node.reading_frame_index != self.reading_frame_index
            for e in self.in_edges)

    def has_in_subgraph(self) -> bool:
        """ check if it has any in node from different reading frame """
        return any(e.in_node.subgraph_id != self.subgraph_id
            for e in self.in_edges)

    def is_bridge_to_subgraph(self) -> bool:
        """ check if it is a bridge node to a subgraph """
        return any(e.in_node.subgraph_id != self.subgraph_id for e in self.in_edges)

    def is_subgraph_bridge(self, out_node:TVGNode) -> bool:
        """ check if it is a subgraph bridge node """
        return self.subgraph_id != out_node.subgraph_id

    def is_subgraph_end(self) -> bool:
        """ check if is the end of a subgraph """
        return len(self.get_out_nodes()) > 0 \
            and all(x.level < self.level for x in self.get_out_nodes())

    def is_orf_bridge(self, out_node:TVGNode) -> bool:
        """ check if it is a orf bridge node """
        return self.reading_frame_index != out_node.reading_frame_index

    def is_out_orf_bridge(self) -> bool:
        """ check if it is an ORF out bridge node """
        return any(x.reading_frame_index != self.reading_frame_index
            for x in self.get_out_nodes())

    def has_bridge_from_reading_frame(self, other_reading_frame_index:int):
        """ Check if it has a in bridge node from another reading frame """
        return any(x.reading_frame_index == other_reading_frame_index\
            for x in self.get_in_bridges())

    def has_ref_position(self, i:int) -> bool:
        """ Check if the node has a given reference position """
        return any(i in x.ref for x in self.seq.locations)

    def get_in_bridges(self) -> List[TVGNode]:
        """ Get all in bridge nodes """
        in_bridges = []
        for edge in self.in_edges:
            in_node = edge.in_node
            if any(v.variant.is_frameshifting() for v in in_node.variants):
                in_bridges.append(in_node)
        return in_bridges

    def is_exclusively_outbond_of(self, other:TVGNode) -> bool:
        """ Checks if the node is exclusively outbond with the other.
        Exclusive binding means the upstream node has only 1 outbond edge,
        and the downstream node has only inbond edge."""
        if not self.is_inbond_of(other):
            return False
        return len(self.out_edges) == 1 and len(other.in_edges) == 1

    def has_exclusive_outbond_node(self) -> bool:
        """ The given node has exclusive outbond node """
        return len(self.out_edges) == 1 and \
            len(self.get_out_nodes()[0].in_edges) == 1

    def has_exclusive_inbond_node(self) -> bool:
        """ The given node has exclusive inbond node """
        return len(self.in_edges) == 1 and \
            len(self.get_in_nodes()[0].out_edges) == 1

    def get_reference_next(self) -> TVGNode:
        """ Get the next node of which the edge is reference (not variant
        or cleavage). """
        if not self.out_edges:
            return None
        if len(self.out_edges) == 1:
            return list(self.out_edges)[0].out_node
        ref_node = None
        for edge in self.out_edges:
            if edge.type in ['reference', 'variant_end']:
                if ref_node is None:
                    ref_node = edge.out_node
                elif not ref_node.branch:
                    ref_node = edge.out_node
        if not ref_node:
            raise ValueError('No reference edge was found.')
        return ref_node

    def get_reference_prev(self) -> TVGNode:
        """ Get the previous node of which the edge is reference (not variant
        or cleavage) """
        for edge in self.in_edges:
            if edge.type == 'reference':
                return edge.in_node
        return None

    def get_node_start_reference_index(self) -> int:
        """ Get the referece index of node start """
        if self.seq.locations and self.variants:
            return min(
                self.seq.locations[0].ref.start,
                self.variants[0].variant.location.start
            )

        if self.seq.locations:
            return self.variants[0].ref.start

        return self.variants[0].variant.location.start

    def get_node_end_reference_index(self) -> int:
        """ Get the reference index of node end """
        if self.seq.locations and self.variants:
            return max(
                self.seq.locations[-1].ref.end,
                self.variants[-1].variant.location.end
            )

        if self.seq.locations:
            return self.variants[-1].ref.end

        return self.variants[-1].variant.location.end

    def copy(self) -> TVGNode:
        """ Create a copy of the node """
        variants = []
        for variant in self.variants:
            new_variant = seqvar.VariantRecordWithCoordinate(
                variant=variant.variant,
                location=copy.deepcopy(variant.location)
            )
            variants.append(new_variant)
        return TVGNode(
            seq=copy.deepcopy(self.seq),
            variants=variants,
            branch=self.branch,
            orf=self.orf,
            reading_frame_index=self.reading_frame_index,
            subgraph_id=self.subgraph_id,
            global_variant=self.global_variant,
            level=self.level,
            was_bridge=self.was_bridge
        )

    def deepcopy(self, subgraph_id:str=None, level_increment:int=None) -> TVGNode:
        """ Create a deep copy of the node and all its downstream nodes.
        """
        new_node = self.copy()
        if level_increment is not None:
            new_node.level += level_increment

        queue:Deque[Tuple[TVGNode, TVGNode]] = deque([(self, new_node)])
        visited:Dict[TVGNode, TVGNode] = {}

        while queue:
            source, target = queue.pop()
            for edge in source.out_edges:
                source_out_node = edge.out_node
                visited_this = False
                if source_out_node in visited:
                    new_out_node = visited[source_out_node]
                    visited_this = True
                else:
                    new_out_node = source_out_node.copy()
                    if subgraph_id:
                        new_out_node.subgraph_id = subgraph_id
                        if new_out_node.seq:
                            for loc in new_out_node.seq.locations:
                                loc.ref.seqname = subgraph_id
                        for variant in new_out_node.variants:
                            variant.location.seqname = subgraph_id
                    if level_increment is not None:
                        new_out_node.level += level_increment
                    visited[source_out_node] = new_out_node
                new_edge = TVGEdge(target, new_out_node,
                    _type=edge.type)
                target.out_edges.add(new_edge)
                new_out_node.in_edges.add(new_edge)
                if not visited_this and edge.out_node.out_edges:
                    queue.appendleft((edge.out_node, new_out_node))

        return new_node

    def stringify(self, k:int=None) -> None:
        """ Get a str representation of a subgraph """
        if not k:
            k = float("inf")

        _type = 'alt' if self.variants else 'ref'
        if self.seq:
            n = len(self.seq.seq)
            seq_str = f'{str(self.seq.seq[:k])}...' if n >= k else self.seq.seq
            node = f'{seq_str}|{_type}'
        else:
            node = 'root'

        if not self.out_edges:
            return {node: '$'}

        downstream = {}
        for edge in self.out_edges:
            downstream.update(edge.out_node.stringify(k))

        return {node: downstream}

    def get_orf_start(self, i:int=0) -> int:
        """ Get the ORF start position given the start codon is found at
        position i of the sequence. """
        if self.seq.locations:
            for loc in self.seq.locations:
                if i < loc.query.start:
                    return int((3 - (loc.query.start - i) % 3) % 3 + loc.ref.start)
                if loc.query.start <= i < loc.query.end:
                    return int(i - loc.query.start + loc.ref.start)
        out_node = self.get_reference_next()
        if out_node.seq.locations:
            loc = out_node.seq.locations[0]
            return int(((len(self.seq.seq) + loc.query.start - i) % 3) % 3 + loc.ref.start)
        raise ValueError('Can not find ORF')

    def truncate_left(self, i:int) -> TVGNode:
        """ Truncate the left i nucleotides off. A new node with the left part
        of the sequences and variants associated is returned. The self node
        is updated with only the right part of the sequence and variants. """
        left_seq = self.seq[:i]
        left_variants = []
        right_variants = []

        for variant in self.variants:
            if variant.location.start < i:
                end = min(i, variant.location.end)
                end_offset = variant.location.end_offset \
                    if end == variant.location.end else 0
                new_loc = FeatureLocation(
                    start=variant.location.start,
                    end=end,
                    seqname=variant.location.seqname,
                    reading_frame_index=variant.location.reading_frame_index,
                    start_offset=variant.location.start_offset,
                    end_offset=end_offset
                )
                new_variant = seqvar.VariantRecordWithCoordinate(
                    variant=variant.variant, location=new_loc
                )
                left_variants.append(new_variant)
            if variant.location.end > i:
                start = max(variant.location.start, i)
                start_offset = variant.location.start_offset \
                    if start == variant.location.start else 0
                new_loc = FeatureLocation(
                    start=start,
                    end=variant.location.end,
                    seqname=variant.location.seqname,
                    reading_frame_index=variant.location.reading_frame_index,
                    start_offset=start_offset,
                    end_offset=variant.location.end_offset
                )
                new_variant = seqvar.VariantRecordWithCoordinate(
                    variant=variant.variant, location=new_loc
                )
                new_variant = new_variant.shift(-i)
                right_variants.append(new_variant)

        left_node = self.create_node(
            seq=left_seq,
            variants=left_variants,
            branch=self.branch,
            orf=self.orf,
            reading_frame_index=self.reading_frame_index,
            subgraph_id=self.subgraph_id,
            global_variant=self.global_variant,
            level=self.level
        )

        self.seq = self.seq[i:]
        self.variants = right_variants
        self.branch = False

        return left_node

    def truncate_right(self, i:int) -> TVGNode:
        """ Truncate the right i nucleotides off. A new node with the right
        part of the sequences and variants associated is returned. The self
        node is updated with only the left part of the sequence and variants.
        """
        right_seq = self.seq[i:]
        left_variants = []
        right_variants = []

        for variant in self.variants:
            if variant.location.start < i:
                end = min(i, variant.location.end)
                end_offset = variant.location.end_offset \
                    if end == variant.location.end else 0
                new_loc = FeatureLocation(
                    start=variant.location.start,
                    end=end,
                    seqname=variant.location.seqname,
                    reading_frame_index=variant.location.reading_frame_index,
                    start_offset=variant.location.start_offset,
                    end_offset=end_offset
                )
                new_variant = seqvar.VariantRecordWithCoordinate(
                    variant=variant.variant, location=new_loc
                )
                left_variants.append(new_variant)
            if variant.location.end > i:
                start = max(variant.location.start, i)
                start_offset = variant.location.start_offset \
                    if start == variant.location.start else 0
                new_loc = FeatureLocation(
                    start=start,
                    end=variant.location.end,
                    seqname=variant.location.seqname,
                    reading_frame_index=variant.location.reading_frame_index,
                    start_offset=start_offset,
                    end_offset=variant.location.end_offset
                )
                new_variant = seqvar.VariantRecordWithCoordinate(
                    variant=variant.variant, location=new_loc
                )
                new_variant = new_variant.shift(-i)
                right_variants.append(new_variant)

        right_node = self.create_node(
            seq=right_seq,
            variants=right_variants,
            branch=False,
            orf=self.orf,
            reading_frame_index=self.reading_frame_index,
            subgraph_id=self.subgraph_id,
            level=self.level,
            was_bridge=self.was_bridge
        )

        self.seq = self.seq[:i]
        self.variants = left_variants

        return right_node

    def append_left(self, other:TVGNode) -> None:
        """ Combine the other node the the left. """
        self.seq = other.seq + self.seq

        variants = copy.copy(other.variants)
        for v_r in self.variants:
            v_l = None
            should_combine_variants = False

            for v_l in other.variants:
                if v_r.location.start != 0:
                    break
                if v_l.location.end != len(other.seq.seq):
                    continue
                if v_l.variant == v_r.variant:
                    should_combine_variants = False
                    break

            if should_combine_variants and v_l:
                v_l.location = FeatureLocation(
                    start=v_l.location.start,
                    end=v_r.location.end + len(other.seq.seq),
                    reading_frame_index=v_l.location.reading_frame_index,
                    start_offset=v_l.location.start_offset,
                    end_offset=v_r.location.end_offset
                )
            else:
                variants.append(v_r.shift(len(other.seq.seq)))
        self.variants = variants

    def append_right(self, other:TVGNode) -> None:
        """ Combine the other node the the right. """
        new_seq = self.seq + other.seq

        for v_r in other.variants:
            v_l = None
            should_combine_variants = False
            for v_l in self.variants:
                if v_r.location.start != 0:
                    break
                if v_l.location.end != len(self.seq.seq):
                    continue
                if v_l.variant == v_r.variant \
                        and v_l.variant.is_real_fusion == v_r.variant.is_real_fusion:
                    should_combine_variants = True
                    break

            if should_combine_variants and v_l:
                v_l.location = FeatureLocation(
                    start=v_l.location.start,
                    end=v_r.location.end + len(self.seq.seq),
                    seqname=v_l.location.seqname,
                    reading_frame_index=v_l.location.reading_frame_index,
                    start_offset=v_l.location.start_offset,
                    end_offset=v_r.location.end_offset
                )
            else:
                self.variants.append(v_r.shift(len(self.seq.seq)))

        self.seq = new_seq

    def translate(self) -> PVGNode:
        """ translate to a PVGNode """
        if not self.out_edges:
            seq = self.seq[:len(self.seq) - len(self.seq) % 3].translate()
        else:
            seq = self.seq.translate()

        locations = []
        for loc in self.seq.locations:
            query_start = math.floor(loc.query.start/ 3)
            query_end = math.ceil(loc.query.end / 3)
            query_start_offset = loc.query.start - query_start * 3
            query_end_offset = query_end * 3 - loc.query.end
            query = FeatureLocation(
                start=query_start, end=query_end,
                reading_frame_index=loc.query.reading_frame_index,
                start_offset=query_start_offset,
                end_offset=query_end_offset
            )
            dna_query_codon_start = query_start * 3
            dna_ref_codon_start = loc.ref.start - (loc.query.start - dna_query_codon_start)
            ref_start = math.floor(dna_ref_codon_start / 3)
            dna_query_codon_end = query_end * 3
            dna_ref_codon_end = loc.ref.end + (dna_query_codon_end - loc.query.end)
            ref_end = ref_start + len(query)
            ref_start_offset = dna_ref_codon_start - ref_start * 3
            ref_end_offset = dna_ref_codon_end - ref_end * 3
            ref = FeatureLocation(
                start=ref_start, end=ref_end, seqname=loc.ref.seqname,
                start_offset=ref_start_offset, end_offset=ref_end_offset
            )
            locations.append(MatchedLocation(query=query, ref=ref))

        seq.__class__ = aa.AminoAcidSeqRecordWithCoordinates
        seq.locations = locations
        seq.orf = self.orf

        # translate the dna variant location to peptide coordinates.
        variants = [v.to_protein_coordinates() for v in self.variants]

        return PVGNode(
            seq=seq,
            variants=variants,
            orf=[None, None],
            reading_frame_index=self.reading_frame_index,
            subgraph_id=self.subgraph_id,
            level=self.level
        )

    def get_ith_variant_var_aa(self, i:int) -> Seq:
        """ Get the variant amino acid sequence of the ith variant """
        v = self.variants[i]
        loc = v.location
        lhs = loc.start - loc.start % 3
        rhs = loc.end + (3 - loc.end % 3) % 3
        rhs = min(rhs, len(self.seq.seq))
        seq = self.seq.seq[lhs:rhs]
        seq = seq[:len(seq) - len(seq) % 3]
        return seq.translate(to_stop=False)

    def get_ith_variant_ref_aa(self, i:int, tx_seq:Seq) -> Seq:
        """ Get the reference amino acid sequence of the ith variant. """
        v = self.variants[i]
        if v.variant.type == 'Insertion':
            left_offset = (v.location.start - 1) % 3
        else:
            left_offset = v.location.start % 3
        lhs = v.variant.location.start - left_offset

        seq = Seq('')
        if v.variant.type in ['Deletion', 'Substitution']:
            seq += tx_seq[v.variant.location.start:v.variant.location.end]
        else:
            seq += Seq(v.variant.ref)

        if left_offset > 0:
            seq = tx_seq[lhs:v.variant.location.start] + seq

        rhs = v.variant.location.end + (3 - len(seq) % 3) % 3
        rhs = min(rhs, len(tx_seq))

        if rhs > v.variant.location.end:
            seq += tx_seq[v.variant.location.end:rhs]
        seq = seq[:len(seq) - len(seq) % 3]
        return seq.translate(to_stop=False)

    def get_ref_sequence(self, tx_seq:Seq) -> Seq:
        """ Get the reference sequence """
        if not self.variants:
            return self.seq.seq

        seq = self.seq.seq[:self.variants[0].location.start]
        for i, variant in enumerate(self.variants):
            if variant.variant.type in ['Deletion', 'Substitution']:
                seq += tx_seq[variant.variant.location.start:variant.variant.location.end]
            elif variant.variant.type != 'Insertion':
                seq += variant.variant.ref
            if i + 1 >= len(self.variants):
                seq += self.seq.seq[variant.location.end:]
            else:
                next_var = self.variants[i + 1]
                seq += self.seq.seq[variant.location.end:next_var.location.start]

        return seq

    def check_stop_altering(self, tx_seq:Seq, cds_end:int=None):
        """ Checks if any variant is stop altering """
        if not self.variants:
            return

        if cds_end:
            stop_codon = FeatureLocation(start=cds_end, end=cds_end + 3)
            if not self.get_node_start_reference_index() <= cds_end < \
                    self.get_node_end_reference_index():
                return
            for variant in self.variants:
                if stop_codon.overlaps(variant.variant.location):
                    variant.is_stop_altering = True
            return

        if self.global_variant and self.global_variant.is_fusion():
            return

        for i,v in enumerate(self.variants):
            if v.variant.type in {'Fusion', 'circRNA'}:
                continue
            ref_aa = self.get_ith_variant_ref_aa(i, tx_seq)
            var_aa = self.get_ith_variant_var_aa(i)
            v.is_silent = v.variant.is_snv() and ref_aa == var_aa
            v.is_stop_altering = \
                (v.variant.is_snv() and ref_aa == '*' and var_aa != '*') \
                or (v.variant.is_insertion() and ref_aa =='*'
                    and not var_aa.startswith('*')) \
                or (v.variant.is_deletion() and '*' in ref_aa
                    and not (ref_aa.startswith('*') and var_aa.startswith('*'))) \
                or (v.variant.type in ['Substitution', 'MNV']
                    and '*' in ref_aa and ref_aa != var_aa)

    def is_less_mutated_than(self, other:TVGNode) -> bool:
        """ Checks if this node has less mutation than the other """
        self_vars = [v for v in self.variants if v.variant is not self.global_variant]
        other_vars = [v for v in other.variants if v.variant is not other.global_variant]
        if self.level != other.level:
            return self.level < other.level

        if len(self_vars) != len(other_vars):
            return len(self_vars) < len(other_vars)

        for x, y in zip(self_vars, other_vars):
            if x.location != y.location:
                return x.location < y.location

        for x, y in zip(self_vars, other_vars):
            # SNV is higher than RES
            if x.variant.type == 'SNV' and y.variant.type == 'RNAEditingSite':
                return True
            if x.variant.type == 'RNAEditingSite' and y.variant.type == 'SNV':
                return False
            # Otherwise don't really care about the order, just to make sure
            # the result is reproducible
            if x.variant.type != y.variant.type:
                return x.variant.type > y.variant.type

            if x.variant.alt != y.variant.type:
                return x.variant.alt > y.variant.type
        return True

    def is_inframe_subgraph(self, start:TVGNode, end:TVGNode) -> bool:
        """ Check whether it is a inframe subgraph, i.e. its a subgraph node
        and its upstream and downstream are in the same reading frame and
        are also from the main graph. """
        if len(self.in_edges) != 1 or len(self.out_edges) != 1:
            return False

        upstream = self.get_in_nodes()[0]
        downstream = self.get_out_nodes()[0]

        return downstream.subgraph_id == start.subgraph_id \
            and upstream.subgraph_id == start.subgraph_id \
            and downstream.reading_frame_index == start.reading_frame_index \
            and upstream.reading_frame_index == start.reading_frame_index \
            and (upstream.seq.seq == ''
                or (upstream.seq.locations[0].ref > start.seq.locations[0].ref)) \
            and downstream.seq.locations[0].ref < end.seq.locations[0].ref

    def get_selenocysteine_positions(self, selenocysteines:List[FeatureLocation]
            ) -> List[int]:
        """ Find selenocysteine position from the sequence """
        positions = []
        for sec in selenocysteines:
            query_i = self.seq.get_query_index(sec.start)
            if query_i == -1:
                continue
            if query_i % 3 != 0:
                continue
            sec_query = FeatureLocation(start=query_i, end=query_i + 3)
            if any(v.location.overlaps(sec_query) for v in self.variants):
                continue
            positions.append(query_i)
        return positions



""" Varaint Peptide Table """
from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Set, IO, List, Tuple
from Bio import SeqUtils
from Bio.SeqIO import FastaIO
from moPepGen import VARIANT_PEPTIDE_SOURCE_DELIMITER, aa
from moPepGen.SeqFeature import FeatureLocation
from moPepGen.svgraph.VariantPeptideDict import AnnotatedPeptideLabel, PeptideSegment


if TYPE_CHECKING:
    from pathlib import Path
    from Bio.Seq import Seq
    from moPepGen.params import CleavageParams

VARIANT_PEPTIDE_TABLE_HEADERS = [
    'sequence', 'header', 'subsequence', 'start', 'end', 'feature_type', 'feature_id',
    'ref_start', 'ref_end', 'start_offset', 'end_offset', 'variant'
]

class VariantPeptideTable:
    """ Variant Peptide Segment Annotation """
    def __init__(self, handle:IO[str], index:Dict[Seq, List[Tuple[int,int]]]=None):
        self.handle = handle
        self.index = index or {}
        self.header_delimeter = VARIANT_PEPTIDE_SOURCE_DELIMITER

    def write_header(self):
        """ Write header """
        self.handle.write('#' + '\t'.join(VARIANT_PEPTIDE_TABLE_HEADERS) + '\n')

    def is_valid(self, seq:Seq, canonical_peptides:Set[str],
            cleavage_params:CleavageParams=None):
        """ Check if the peptide is valid """
        min_mw = cleavage_params.min_mw
        min_length = cleavage_params.min_length
        max_length = cleavage_params.max_length
        if SeqUtils.molecular_weight(seq, 'protein') < min_mw:
            return False
        if len(seq) < min_length or len(seq) > max_length:
            return False
        if str(seq) in canonical_peptides:
            return False
        return True

    def add_peptide(self, seq:Seq, peptide_anno:AnnotatedPeptideLabel):
        """ Write """
        start = self.handle.tell()
        for seg in peptide_anno.segments:
            subseq = str(seq[seg.query.start:seg.query.end])
            line = f"{str(seq)}\t{peptide_anno.label}\t{subseq}\t{seg.to_line()}\n"
            self.handle.write(line)
        end = self.handle.tell()
        cur = (start, end)
        if seq in self.index:
            self.index[seq].append(cur)
        else:
            self.index[seq] = [cur]

    def load_peptide(self, seq:Seq):
        """ Load peptide from table """
        labels = set()
        for start, end in self.index[seq]:
            self.handle.seek(start, 0)
            buffer:str = self.handle.read(end - start)
            for line in buffer.rstrip().split('\n'):
                fields = line.split('\t')
                if seq != fields[0]:
                    raise ValueError(
                        f"Peptide ({seq}) do not match with table record ({fields[0]})."
                    )
                labels.add(fields[1])
        label = self.header_delimeter.join(labels)

        return aa.AminoAcidSeqRecord(
            seq=seq,
            description=label,
            name=label
        )

    def load_peptide_annotation(self, seq):
        """ Load peptide annotation """
        anno:Dict[str, AnnotatedPeptideLabel] = {}
        for start, end in self.index[seq]:
            self.handle.seek(start)
            buffer:str = self.handle.read(end - start)
            for line in buffer.rstrip().split('\n'):
                fields = dict(zip(VARIANT_PEPTIDE_TABLE_HEADERS, line.split('\t')))
                if seq != fields['sequence']:
                    raise ValueError(
                        f"Peptide ({seq}) does not match with table record ({fields[0]})."
                    )
                header = fields['header']
                label = anno.setdefault(header, AnnotatedPeptideLabel(label=header, segments=[]))
                query = FeatureLocation(
                    start=int(fields['start']),
                    end=int(fields['end']),
                    start_offset=int(fields['start_offset']),
                    end_offset=int(fields['end_offset'])
                )
                feature_type = fields['feature_type'] if fields['feature_type'] != '.' else None
                feature_id = fields['feature_id'] if fields['feature_id'] != '.' else None

                if fields['ref_start'] != '.':
                    ref_start_raw = int(fields['ref_start'])
                    ref_start = int(ref_start_raw / 3)
                    ref_start_offset = ref_start_raw - ref_start * 3
                    ref_end_raw = int(fields['ref_end'])
                    ref_end = int(ref_end_raw / 3)
                    ref_end_offset = ref_end_raw - ref_end * 3

                    ref = FeatureLocation(
                        start=ref_start,
                        end=ref_end,
                        start_offset=ref_start_offset,
                        end_offset=ref_end_offset
                    )
                else:
                    ref = None
                variant = fields['variant'] if fields['variant'] != '.' else None
                seg = PeptideSegment(
                    query=query, ref=ref, feature_type=feature_type,
                    feature_id=feature_id, variant_id=variant
                )
                label.segments.append(seg)
        return anno

    def write_fasta(self, path:Path):
        """ write fasta """
        with open(path, 'wt') as handle:
            record2title = lambda x: x.description
            writer = FastaIO.FastaWriter(handle, record2title=record2title)
            for seq in self.index:
                peptide = self.load_peptide(seq)
                writer.write_record(peptide)

    def generate_index(self):
        """ Generate index from peptide table file. """
        self.handle.seek(0)
        while True:
            cur_start = self.handle.tell()
            line = self.handle.readline()
            cur_end = cur_start + len(line)
            if not line:
                break
            if line.startswith('#'):
                continue
            seq = line.split('\t')[0]
            indices = self.index.setdefault(seq, [])
            if indices and indices[-1][1] == cur_start:
                indices[-1] = (indices[-1][0], cur_end)
            else:
                indices.append((cur_start, cur_end))



""" Module for CircRNA """
from __future__ import annotations
from typing import List, TYPE_CHECKING
from moPepGen.SeqFeature import SeqFeature, FeatureLocation


if TYPE_CHECKING:
    from moPepGen.dna import DNASeqRecordWithCoordinates
    from moPepGen.gtf import GeneAnnotationModel

class CircRNAModel():
    """
    Attributes:
        transcript_id (str)
        fragments (List[SeqFeature])
        intron (List[int])
        id (str)
        gene_id (str)
        gene_name (str)
        gene_locations (List[SeqFeature])
    """
    def __init__(self, transcript_id:str, fragments:List[SeqFeature],
            intron:List[int], _id:str, gene_id:str, gene_name:str,
            genomic_location:str=''):
        """ Constructor """
        self.gene_id = gene_id
        self.fragments = fragments
        self.intron = intron
        self.id = _id
        self.transcript_id = transcript_id
        self.gene_name = gene_name
        self.gene_locations = []
        self.genomic_position = genomic_location

    def get_gene_coordinates(self, gene:GeneAnnotationModel) -> None:
        """ Get the coordinates of the gene """
        features:List[SeqFeature] = []
        for i,fragment in enumerate(self.fragments):
            start = fragment.location.start
            end = fragment.location.end
            feature = SeqFeature(
                chrom=gene.id,
                location=FeatureLocation(
                    seqname=gene.id, start=start,
                    end=end, strand=gene.strand
                ),
                attributes={},
                type='intron' if i + 1 in self.intron else 'exon'
            )
            features.append(feature)
        self.gene_locations = features

    def get_circ_rna_sequence(self, seq:DNASeqRecordWithCoordinates):
        """ Get the DNA sequence of the circRNA.

        Args:
            seq (DNASeqRecordWithCoordinates): The DNA sequence of the
                gene where the circRNA comes from.
        """
        circ = None
        for fragment in sorted(self.fragments):
            new_seq = seq[int(fragment.location.start):int(fragment.location.end)]
            circ = circ + new_seq if circ else new_seq
        return circ

    def to_string(self) -> str:
        """ Convert to a string """
        gene_id = self.gene_id
        start = int(self.fragments[0].location.start)
        offset, length = [], []
        for fragment in self.fragments:
            offset.append(str(fragment.location.start - start))
            length.append(str(fragment.location.end - fragment.location.start))
        start = str(start)
        offset = ','.join(offset)
        length = ','.join(length)
        intron = ','.join([str(x) for x in self.intron])
        circ_id = self.id
        tx_id = self.transcript_id
        gene_name = self.gene_name
        info = f'OFFSET={offset};LENGTH={length};INTRON={intron};' +\
            f'TRANSCRIPT_ID={tx_id};GENE_SYMBOL={gene_name};' +\
            f'GENOMIC_POSITION={self.genomic_position}'
        return '\t'.join([gene_id, start, circ_id, '.', '.', '.', '.', info])



""" Module for moPepGen seqvar IO """
from typing import Dict, Iterable, Tuple, Union, IO
from pathlib import Path
import tempfile
from moPepGen import GVF_HEADER, constant
from moPepGen.seqvar.GVFMetadata import GVFMetadata
from moPepGen.seqvar.VariantRecord import VariantRecord
from moPepGen.SeqFeature import FeatureLocation


T = Union[str, IO, Path]

def parse(handle:T) -> Iterable[VariantRecord]:
    """ Parse a GVF file.

    Args:
        path (str): Path to the moPepGen variant file.
        type (str):
    Return:
        An iterable of seq
    """
    if isinstance(handle, (str, Path)):
        with open(handle, 'r') as stream:
            for record in iterate(stream):
                yield record
    else:
        for record in iterate(handle):
            yield record

def iterate(handle:IO) -> Iterable[VariantRecord]:
    """ Iterate through a file-handle of a GVF file and returns an iterator
    of VariantRecord objects.

    Args:
        handle (IO): File-handle like object. Usually the TextIOWrapper
            returned by the open statement.
        it (str): An iterator if the iterating already starts.

    Example:
        >>> with open('test.gvf', 'rt') as handle:
                for record in iterate(handle):
                    print(record)
                    break
        <VariantRecord>
    """
    for line in handle:
        if line.startswith('#'):
            continue
        yield line_to_variant_record(line)


def parse_attrs(info:str) -> Dict[str, Union[str,int]]:
    """ parse the attributes of a GVF record. """
    attrs = {}
    for field in info.split(';'):
        key, val = field.split('=')
        val = val.strip('"')
        if key in constant.ATTRS_POSITION:
            val = str(int(val) - 1)
        attrs[key] = val
    return attrs

def line_to_variant_record(line:str) -> VariantRecord:
    """ Convert line from GVF file to VariantRecord """
    fields = line.rstrip().split('\t')
    gene_id = fields[0]
    start=int(fields[1]) - 1
    ref = fields[3]
    alt = fields[4]
    attrs = parse_attrs(fields[7])

    if not alt.startswith('<'):
        end = start + len(ref)
        if len(ref) == len(alt) == 1:
            _type = 'SNV'
        elif len(ref) == 1 or len(alt) == 1:
            _type = 'INDEL'
        else:
            _type = 'MNV'
    elif alt == '<FUSION>':
        end = start + 1
        _type = 'Fusion'
    elif alt == '<DEL>':
        end = int(attrs['END'])
        _type = 'Deletion'
    elif alt == '<INS>':
        end = start + 1
        _type = 'Insertion'
    elif alt == '<SUB>':
        end = int(attrs['END'])
        _type = 'Substitution'
    else:
        raise ValueError('Alt type not supported.')

    _id = fields[2]

    return VariantRecord(
        location=FeatureLocation(
            seqname=gene_id,
            start=start,
            end=end
        ),
        ref=ref,
        alt=alt,
        _type=_type,
        _id=_id,
        attrs=attrs
    )

def parse_label(handle:IO) -> Tuple[str, str, str]:
    """ parse only the transcirpt ID and variant label """
    line:str
    for line in handle:
        if line.startswith('#'):
            continue
        fields = line.rstrip().split('\t')
        gene_id = fields[0]
        label = fields[2]
        attrs = parse_attrs(fields[7])
        tx_id = attrs['TRANSCRIPT_ID']
        yield gene_id, tx_id, label

def write(variants:Iterable[VariantRecord], output_path:str,
        metadata:GVFMetadata):
    """ Write variants to an out handle.

    Args:
        variants (Iterable[seqvar.VariantRecord]): An iterable of variant
            records to write.
        handle (IO): The destination handle to write out.
        mode (str): If 'w', the header will be written, otherwise not.
    """
    with tempfile.TemporaryFile(mode='w+t') as temp_file:
        temp_file.write('#' + '\t'.join(GVF_HEADER) + '\n')
        for record in variants:
            line = record.to_string()
            temp_file.write(line + '\n')
            metadata.add_info(record.type)
        with open(output_path, 'w') as out_file:
            for line in metadata.to_strings():
                out_file.write(line + '\n')

            temp_file.seek(0)
            for line in temp_file:
                out_file.write(line)



""" Module for GTF IO """
from __future__ import annotations
from typing import IO, Union, Iterable, TYPE_CHECKING
from Bio.SeqIO.Interfaces import SequenceIterator
from moPepGen.SeqFeature import FeatureLocation
from .GTFSeqFeature import GTFSeqFeature


if TYPE_CHECKING:
    from moPepGen.gtf import GenomicAnnotation

class GtfIterator(SequenceIterator):
    """ GTF Iterator """
    def __init__(self, source:Union[IO, str], mode='t'):
        """ Constructor """
        super().__init__(source=source, mode=mode, fmt='GTF')

    def parse(self, handle:IO[str]) -> Iterable[GTFSeqFeature]:
        """ parse
        """
        records = self.iterate(handle)
        return records

    @staticmethod
    def iterate(handle:IO[str]) -> Iterable[GTFSeqFeature]:
        """ Iterate through a GTF file and yield a record each time. """
        for line in handle:
            if line.startswith('#'):
                continue
            line = line.rstrip()
            record = line_to_seq_feature(line)

            yield record

def line_to_seq_feature(line:str) -> GTFSeqFeature:
    """ Conver line to SeqFeature """
    line = line.rstrip()
    fields = line.split('\t')

    try:
        strand={'+':1, '-':-1, '?':0}[fields[6]]
    except KeyError:
        strand=None

    attributes = {}
    attributes_to_keep = ['gene_id', 'transcript_id', 'protein_id',
        'gene_name', 'gene_type', 'gene_biotype', 'tag', 'is_protein_coding']
    attribute_list = [field.strip().split(' ', 1) for field in \
        fields[8].rstrip(';').split(';')]

    for key,val in attribute_list:
        if key not in attributes_to_keep:
            continue
        val = val.strip('"')
        if key == 'tag':
            if key not in attributes:
                attributes[key] = []
            attributes[key].append(val)
        else:
            attributes[key] = val

    location = FeatureLocation(
        seqname=fields[0],
        start=int(fields[3])-1,
        end=int(fields[4]),
        strand=strand,
    )

    frame = None if fields[7] == '.' else int(fields[7])

    return GTFSeqFeature(
        chrom=fields[0],
        attributes=attributes,
        location=location,
        type=fields[2],
        frame=frame
    )

def parse(handle:Union[IO[str], str]) -> GtfIterator:
    """ Parser for GTF files.
    """
    return GtfIterator(handle)

def to_gtf_record(record:GTFSeqFeature, is_protein_coding:bool=None) -> str:
    """ Convert a SeqFeature object to a GTF record """
    if record.strand == 1:
        strand = '+'
    elif record.strand == -1:
        strand = '-'
    else:
        strand = '.'

    attrs = ""
    for key, val in record.attributes.items():
        if isinstance(val, list):
            for vali in val:
                attrs += f" {key} {vali};"
        else:
            attrs += f" {key} {val};"

    if is_protein_coding is not None:
        is_protein_coding = 'true' if is_protein_coding is True else 'false'
        attrs += f" is_protein_coding {is_protein_coding};"

    frame = '.' if record.frame is None else str(record.frame)
    record_data = [
        record.chrom, '.', record.type, str(int(record.location.start)+1),
        str(int(record.location.end)), '.', strand, frame, attrs
    ]
    return '\t'.join(record_data)

def write(handle:IO, anno:GenomicAnnotation) -> None:
    """ Write an GenomicAnnotation as a GTF file. """
    for gene_model in anno.genes.values():
        handle.write(to_gtf_record(gene_model) + '\n')
        for tx_id in gene_model.transcripts:
            tx_model = anno.transcripts[tx_id]
            record = to_gtf_record(tx_model.transcript, tx_model.is_protein_coding)
            handle.write(record + '\n')
            records = tx_model.cds + tx_model.exon
            records.sort()
            records.extend(tx_model.utr)
            records = tx_model.selenocysteine + records
            for record in records:
                handle.write(to_gtf_record(record) + '\n')



""" Module for fuzz test """
from __future__ import annotations
import argparse
import copy
from contextlib import redirect_stdout
from datetime import datetime, timedelta
from enum import Enum
import json
from pathlib import Path
import random
import shutil
import traceback
from typing import Iterable, List, Union, Tuple, IO
import sys
import uuid
from pathos.pools import ParallelPool
from Bio import SeqIO
from moPepGen import ERROR_INDEX_IN_INTRON, fake, seqvar, circ, dna, gtf, aa, get_logger
from moPepGen.seqvar.VariantRecord import VariantRecord
from moPepGen.circ import CircRNAModel
from moPepGen.util.common import load_references
from moPepGen.cli.common import add_args_cleavage, generate_metadata, \
    print_help_if_missing_args, add_args_debug_level
from moPepGen.cli import call_variant_peptide
from moPepGen.gtf import GtfIO
from . import brute_force


class FuzzRecordStatus(str, Enum):
    """ Enumeration for fuzz record status. """
    succeeded = 'SUCCEEDED'
    failed = 'FAILED'

    def __str__(self) -> str:
        """ str """
        return self.value

# pylint: disable=W0212
def parse_args(subparsers:argparse._SubParsersAction
        ) -> argparse.ArgumentParser:
    """ Arguments for the fuzzTest subcommand """
    parser:argparse.ArgumentParser = subparsers.add_parser(
        name='fuzzTest',
        help='Fuzz test that randomly generates variants data and validate'
        ' the graph algorithm to brute force.'
    )
    parser.add_argument(
        '--reference-dir',
        type=Path,
        help='Directory to reference files. Noted this reference directory must'
        ' contain reference files generated by the downsampleReference command'
        ' and is different from the --index-dir for many moPepGen commands.',
        metavar = '<directory>',
        default=None
    )
    parser.add_argument(
        '--tx-id',
        type=str,
        help='Transcript ID',
        metavar='<value>'
    )
    parser.add_argument(
        '-s', '--max-size',
        type=int,
        help='Maximal number of nucleotides being inserted or deleted. Only'
        ' SNVs are generated when setting it to 0.',
        metavar='<number>'
    )
    parser.add_argument(
        '--max-variants',
        type=int,
        help='Max number of variants',
        metavar='<number>'
    )
    parser.add_argument(
        '--min-variants',
        type=int,
        help='Min number of variants',
        default=1,
        metavar='<number>'
    )
    parser.add_argument(
        '--snv-frac',
        type=float,
        default=1,
        help='Fraction of SNVs to simulate over INDEL in any test case. For'
        ' example, `--snv--frac 0.5` will have roughly 50% SNVs and 50% INDELs.'
        ' This has no effect on fusion, circRNA, or altSplice.'
    )
    parser.add_argument(
        '--snv-only-frac',
        type=float,
        default=1,
        help='Fraction of test cases to be SNV only (no INDEL). For example,'
        ' `--snv-only-frac 0.5` will have roughly 50% test cases to have only'
        ' SNV without any INDELs. This has no effect on fusion, circRNA, or altSplice.'
    )
    parser.add_argument(
        '--fusion-frac',
        type=float,
        help='Fraction of test cases to have fusion.',
        default=0
    )
    parser.add_argument(
        '--circ-rna-frac',
        type=float,
        help='Fraction of test cases to have circRNA.',
        default=0
    )
    parser.add_argument(
        '--ci-ratio',
        type=float,
        help='Fraction of ciRNA to generate.',
        default=0.2,
        metavar='<value>'
    )
    parser.add_argument(
        '--alt-splice-frac',
        type=float,
        help='Fraction of test cases to have alternative splicing.',
        default=0
    )
    parser.add_argument(
        '-i', '--n-iter',
        type=int,
        help='Max number of iterations',
        metavar='<number>'
    )
    parser.add_argument(
        '-t', '--temp-dir',
        type=Path,
        help='Temporary directory',
        metavar='<path>'
    )
    parser.add_argument(
        '-o', '--output-dir',
        type=Path,
        help='Output directory',
        metavar='<path>'
    )
    parser.add_argument(
        '--exonic-only',
        action='store_true',
        help='Create only exonic variants',
        default=False
    )
    parser.add_argument(
        '--keep-succeeded',
        action='store_true',
        help='Keep fuzz test files of the succeeded cases.'
    )
    parser.add_argument(
        '--save-graph',
        action='store_true',
        help='Save graph data.'
    )
    parser.add_argument(
        '--seed',
        type=int,
        default=None,
        help='Random seed to set. If not specified, a random number will be generated.'
    )
    parser.add_argument(
        '--nthreads',
        type=int,
        default=1,
        help='Number of threads'
    )
    add_args_cleavage(parser)
    add_args_debug_level(parser)
    parser.set_defaults(func=main)
    print_help_if_missing_args(parser)
    return parser

def plain_logger(msg:str):
    """ log message to stdout """
    print(msg, flush=True, file=sys.stdout)

class FuzzRecord():
    """ Record of the fuzz test result for a single test case """
    def __init__(self, _id:str, status:str, submitted:datetime,
            completed:datetime, work_dir:Path=None, n_var:int=None,
            n_exonic:int=None, n_snv:int=0, n_indel:int=0, n_fusion:int=0,
            n_circ_rna:int=0, n_alt_splice:int=0, call_variant_start:datetime=None,
            call_variant_end:datetime=None, brute_force_start:datetime=None,
            brute_force_end:datetime=None):
        """ constructor """
        self.id = _id
        self.status = status
        self.submitted = submitted
        self.completed = completed
        self.work_dir = work_dir
        self.n_var = n_var
        self.n_exonic = n_exonic
        self.n_snv = n_snv
        self.n_indel = n_indel
        self.n_fusion = n_fusion
        self.n_circ_rna = n_circ_rna
        self.n_alt_splice = n_alt_splice
        self.n_snv = n_snv
        self.call_variant_start = call_variant_start
        self.call_variant_end = call_variant_end
        self.brute_force_start = brute_force_start
        self.brute_force_end = brute_force_end

    def submit(self):
        """ set the submit timestamp """
        self.submitted = datetime.now()

    def complete(self, status:str):
        """ set the complete timestamp """
        self.completed = datetime.now()
        self.status = status

    def call_variant_starts(self):
        """ set timestamp when callVariant starts """
        self.call_variant_start = datetime.now()

    def call_variant_ends(self):
        """ set timestamp when callVariant ends """
        self.call_variant_end = datetime.now()

    def brute_force_starts(self):
        """ set timestamp when bruteForce starts """
        self.brute_force_start = datetime.now()

    def brute_force_ends(self):
        """ set timestamp when bruteForce ends """
        self.brute_force_end = datetime.now()

    def set_work_dir(self, global_work_dir:Path):
        """ set test case specific work dir """
        self.work_dir = global_work_dir/self.id
        self.work_dir.mkdir(exist_ok=True)

    @classmethod
    def generate_record(cls) -> FuzzRecord:
        """ Generate a fuzz record """
        _id = uuid.uuid4().hex
        record =  cls(
            _id=_id, status=None, submitted=None, completed=None
        )
        record.submit()
        return record

    @classmethod
    def parse(cls, handle:IO) -> Iterable[FuzzRecord]:
        """ parse """
        handle.seek(0)
        header = handle.readline().rstrip().split('\t')
        time_fmt = '%Y-%m-%d %H:%M:%S'
        for line in handle:
            line:str
            fields = line.rstrip().split('\t')
            values = dict(zip(header, fields))

            if values['call_variant_start'] == '-':
                call_variant_start = None
            else:
                call_variant_start = datetime.strptime(values['call_variant_start'], time_fmt)

            if values['call_variant_end'] == '-':
                call_variant_end = None
            else:
                call_variant_end = datetime.strptime(values['call_variant_end'], time_fmt)

            if values['brute_force_start'] == '-':
                brute_force_start = None
            else:
                brute_force_start = datetime.strptime(values['brute_force_start'], time_fmt)

            if values['brute_force_end'] == '-':
                brute_force_end = None
            else:
                brute_force_end = datetime.strptime(values['brute_force_end'], time_fmt)

            yield FuzzRecord(
                _id=values['id'],
                status=values['status'],
                submitted=datetime.strptime(values['submitted'], time_fmt),
                completed=datetime.strptime(values['completed'], time_fmt),
                n_var=int(values['num_var']),
                n_exonic=int(values['num_exonic']),
                n_snv=int(values['n_snv']),
                n_indel=int(values['n_indel']),
                n_fusion=int(values['n_fusion']),
                n_circ_rna=int(values['n_circ_rna']),
                n_alt_splice=int(values['n_alt_splice']),
                call_variant_start=call_variant_start,
                call_variant_end=call_variant_end,
                brute_force_start=brute_force_start,
                brute_force_end=brute_force_end
            )

    @property
    def var_gvf_file(self) -> Path:
        """ File path to the GVF file that contains the faked variants """
        return self.work_dir/'fake_variants.gvf'

    @property
    def circ_gvf_file(self) -> Path:
        """ CircRNA GVF file """
        return self.work_dir/'fake_circ_rna.gvf'

    @property
    def call_variant_fasta(self) -> Path:
        """ FASTA file outputted by callVariant """
        return self.work_dir/'call_variant.fasta'

    @property
    def brute_force_fasta(self) -> Path:
        """ FASTA file outputted by bruteForce """
        return self.work_dir/'brute_force.fasta'

    @property
    def call_variant_only(self) -> Path:
        """ Path to the callVariant only file. """
        return self.work_dir/'call_variant_only.txt'

    @property
    def brute_force_only(self) -> Path:
        """ Path to the bruteForce only file. """
        return self.work_dir/'brute_force_only.txt'

    @property
    def timedelta(self) -> timedelta:
        """ Get time difference from submitted to completed """
        return self.completed - self.submitted

    def to_tsv(self) -> str:
        """ convert to TSV record """
        submitted = self.submitted.strftime('%Y-%m-%d %H:%M:%S') \
            if self.submitted else '-'
        completed = self.completed.strftime('%Y-%m-%d %H:%M:%S') \
            if self.completed else '-'
        call_variant_start = self.call_variant_start.strftime('%Y-%m-%d %H:%M:%S') \
            if self.call_variant_start else '-'
        call_variant_end = self.call_variant_end.strftime('%Y-%m-%d %H:%M:%S') \
            if self.call_variant_end else '-'
        call_variant_time = str(self.call_variant_end - self.call_variant_start) \
            if self.call_variant_start and self.call_variant_end else '-'
        brute_force_start = self.brute_force_start.strftime('%Y-%m-%d %H:%M:%S') \
            if self.brute_force_start else '-'
        brute_force_end = self.brute_force_end.strftime('%Y-%m-%d %H:%M:%S') \
            if self.brute_force_end else '-'
        brute_force_time = str(self.brute_force_end - self.brute_force_start) \
            if self.brute_force_start and self.brute_force_end else '-'
        return '\t'.join([
            self.id, self.status, str(self.n_var), str(self.n_exonic),
            str(self.n_snv), str(self.n_indel), str(self.n_fusion),
            str(self.n_circ_rna), str(self.n_alt_splice), submitted, completed,
            call_variant_start, call_variant_end, call_variant_time,
            brute_force_start, brute_force_end, brute_force_time
        ])

    @staticmethod
    def tsv_header() -> str:
        """ get TSV header """
        return '\t'.join([
            'id', 'status', 'num_var', 'num_exonic', 'n_snv', 'n_indel',
            'n_fusion', 'n_circ_rna', 'n_alt_splice', 'submitted', 'completed',
            'call_variant_start', 'call_variant_end', 'call_variant_time',
            'brute_force_start', 'brute_force_end', 'brute_force_time'
        ])

    @property
    def path_stderr(self) -> Path:
        """ File path to the test case specific stderr """
        return self.work_dir/'fuzzing.err'

    @property
    def path_stdout(self) -> Path:
        """ File path to the test case specific stdout """
        return self.work_dir/'fuzzing.out'

    def is_succeeded(self) -> bool:
        """ check if the test case was successful """
        return self.status == FuzzRecordStatus.succeeded

class FuzzTestCase():
    """ This models a single fuzz test case. It creates faked variants, save into
    a GVF file, and then calls moPepGen's callVariant and bruteForce in
    sequence. """
    def __init__(self, config:FuzzTestConfig, record:FuzzRecord=None):
        """ constructor """
        self.config = config
        self.record = record or FuzzRecord.generate_record()
        if not record:
            self.record.set_work_dir(self.config.temp_dir)

    def run(self):
        """ run the current test case """
        logger = get_logger()
        if self.config.seed is None:
            seed = random.randint(1, 1_000_000_000_000)
        else:
            seed = self.config.seed
        random.seed(seed)
        try:
            with open(self.record.path_stdout, 'w') as os_handle:
                with redirect_stdout(os_handle):
                    logger.info("[ fuzzTest ] Random seed: %i", seed)
                    self.fix_fusion_and_circ()
                    if self.config.ref_dir is None:
                        logger.info('[ fuzzTest ] Creating random reference.')
                        genome, anno, proteome = self.create_fake_reference()
                        self.write_fake_reference(genome, anno, proteome)
                        self.config.ref_dir = self.config.temp_dir/self.record.id
                        self.config.tx_id = list(anno.transcripts.keys())[0]
                    with open(self.record.work_dir/'config.json', 'w') as param_handle:
                        json.dump(self.config.to_dict(), param_handle, indent=4)
                    logger.info('Creating random variants.')
                    variants = self.create_variants()
                    self.write_variants(variants)
                    self.record.call_variant_starts()
                    logger.info('Starting callVariant.')
                    self.call_variants()
                    self.record.call_variant_ends()
                    self.record.brute_force_starts()
                    logger.info('Starting bruteForce.')
                    self.brute_force()
                    self.record.brute_force_ends()
                    status = self.assert_equal()
                    self.record.complete(status)
            if status == FuzzRecordStatus.succeeded and not self.config.keep_succeeded:
                shutil.rmtree(self.config.temp_dir/self.record.id)
            else:
                shutil.copytree(
                    self.config.temp_dir/self.record.id,
                    self.config.output_dir/self.record.id
                )
        # pylint: disable=W0703
        except Exception as error:
            self.record.complete(str(FuzzRecordStatus.failed))
            with self.record.path_stderr.open('wt') as handle:
                handle.write(str(error))
                handle.write(traceback.format_exc())
            shutil.copytree(
                self.config.temp_dir/self.record.id,
                self.config.output_dir/self.record.id
            )

    def fix_fusion_and_circ(self):
        """ Fix number of fusion and circRNA """
        if random.choices(
            [True, False],
            weights=[self.config.fusion_frac, 1-self.config.fusion_frac]
        )[0]:
            self.record.n_fusion = 1

        if random.choices(
            [True, False],
            weights=[self.config.circ_rna_frac, 1-self.config.circ_rna_frac]
        )[0]:
            self.record.n_circ_rna = 1

    def create_fake_reference(self
            ) -> Tuple[dna.DNASeqDict, gtf.GenomicAnnotation, aa.AminoAcidSeqDict]:
        """ Create a fake genomic reference, including the genome sequence and
        genomic annotation model, and the translated amino acid sequences. """
        if self.record.n_fusion > 0:
            n_genes = 2
        else:
            n_genes = 1
        genome, anno = fake.fake_genome_and_annotation(n_genes)
        proteome = aa.AminoAcidSeqDict()

        for tx_model in anno.transcripts.values():
            if not tx_model.is_protein_coding:
                continue
            tx_seq = tx_model.get_transcript_sequence(genome[tx_model.transcript.chrom])
            aa_seq = tx_seq[tx_seq.orf.start:tx_seq.orf.end].translate()
            for sec in tx_seq.selenocysteine:
                sec_start = int((sec.start - tx_seq.orf.start) / 3)
                aa_seq = aa_seq[:sec_start] + 'U' + aa_seq[sec_start+1:]
            aa_seq.description = \
                f"{tx_model.protein_id}|{tx_model.transcript_id}|{tx_model.gene_id}|XXX"
            proteome[tx_model.transcript_id] = aa_seq

        return genome, anno, proteome

    def write_fake_reference(self, genome:dna.DNASeqDict, anno:gtf.GenomicAnnotation,
            proteome:aa.AminoAcidSeqDict):
        """ Write the fake genomic reference into files. """
        work_dir = self.config.temp_dir/self.record.id

        with open(work_dir/'annotation.gtf', 'wt') as handle:
            GtfIO.write(handle, anno)

        with open(work_dir/'genome.fasta', 'wt') as handle:
            writer = SeqIO.FastaIO.FastaWriter(handle, record2title=lambda x:x.id)
            for record in genome.values():
                writer.write_record(record)

        with open(work_dir/'proteome.fasta', 'wt') as handle:
            writer = SeqIO.FastaIO.FastaWriter(handle, record2title=lambda x: x.description)
            for record in proteome.values():
                writer.write_record(record)

    def create_variants(self) -> List[Union[VariantRecord,CircRNAModel]]:
        """ create random variant records """
        anno, genome, _ = load_references(
            path_anno=self.config.path_annotation_gtf,
            path_genome=self.config.path_genome_fasta,
            path_proteome=self.config.path_proteome_fasta
        )
        records:List[Union[VariantRecord, CircRNAModel]] = []
        n_variants = random.randint(self.config.min_vairants, self.config.max_variants)
        tx_ids = [self.config.tx_id]

        if self.record.n_fusion > 0:
            if random.choice([True, False]):
                self.record.n_fusion = 1
                record = fake.fake_fusion(anno, genome, self.config.tx_id)
                records.append(record)
                n_variants -= 1
                tx_ids.append(record.attrs['ACCEPTER_TRANSCRIPT_ID'])

        if self.record.n_circ_rna > 0:
            if random.choice([True, False]):
                self.record.n_circ_rna = 1
                record = fake.fake_circ_rna_model(anno, self.config.tx_id,
                    self.config.ci_ratio)
                records.append(record)
                n_variants -= 1

        var_types = ['SNV']
        if random.choices(
            [True, False],
            weights=[self.config.alt_splice_frac, 1-self.config.alt_splice_frac]
        )[0]:
            var_types.append('AltSplicing')
        snv_only = random.choices(
            [True, False],
            weights=[self.config.snv_only_frac, 1-self.config.snv_only_frac]
        )[0]
        for _ in range(n_variants):
            tx_id = random.choice(tx_ids)
            var_type = random.choice(var_types)
            if var_type == 'AltSplicing':
                record = fake.fake_rmats_record(anno, genome, tx_id)
                self.record.n_alt_splice += 1
            elif var_type == 'SNV':
                if snv_only:
                    var_type_sub = 'SNV'
                else:
                    var_type_sub = random.choices(
                        ['SNV', 'INDEL'],
                        weights=[self.config.snv_frac, 1-self.config.snv_frac]
                    )[0]
                record = fake.fake_variant_record(
                    anno=anno, genome=genome, tx_id=tx_id,
                    var_type=var_type_sub,
                    max_size=self.config.max_size,
                    exonic_only=self.config.exonic_only
                )
                if record.type == 'SNV':
                    self.record.n_snv += 1
                else:
                    self.record.n_indel += 1
            else:
                raise ValueError(f"var_type of {var_type} can not be recognized.")
            records.append(record)
        self.record.n_var = n_variants
        self.record.n_exonic = 0
        for variant in records:
            if variant.__class__ is CircRNAModel:
                continue
            try:
                variant.to_transcript_variant(anno, genome, variant.transcript_id)
                self.record.n_exonic += 1
            except ValueError as e:
                if e.args[0] == ERROR_INDEX_IN_INTRON:
                    continue
                raise e

        return records

    def write_variants(self, variants:Iterable[Union[VariantRecord,CircRNAModel]]):
        """ Write variants to file """
        var_records = []
        circ_records = []

        for v in variants:
            if v.__class__ is CircRNAModel:
                circ_records.append(v)
            else:
                var_records.append(v)

        args = argparse.Namespace()
        args.index_dir = None
        args.command = 'parseVEP'
        args.source = ''
        metadata = generate_metadata(args)
        seqvar.io.write(var_records, self.record.var_gvf_file, metadata)

        args.command = 'parseCIRCexplorer'
        metadata = generate_metadata(args)
        if circ_records:
            with open(self.record.circ_gvf_file, 'w') as handle:
                circ.io.write(circ_records, metadata, handle)

    def call_variants(self):
        """ call variants using moPepGen's graph algorithm """
        args = argparse.Namespace()
        args.index_dir = None
        args.command = 'callPeptides'
        args.input_path = [self.record.var_gvf_file]
        if self.record.circ_gvf_file.exists():
            args.input_path.append(self.record.circ_gvf_file)
        args.genome_fasta = self.config.path_genome_fasta
        args.annotation_gtf = self.config.path_annotation_gtf
        args.proteome_fasta = self.config.path_proteome_fasta
        args.reference_source = None
        args.output_path = self.record.call_variant_fasta
        if self.config.save_graph:
            args.graph_output_dir = self.record.work_dir/'graph'
            args.graph_output_dir.mkdir(exist_ok=True)
        else:
            args.graph_output_dir = None
        args.quiet = True
        args.backsplicing_only = False
        args.max_adjacent_as_mnv = 2
        args.selenocysteine_termination = True
        args.w2f_reassignment = True
        args.cleavage_rule = 'trypsin'
        args.cleavage_exception = None
        args.miscleavage = 2
        args.min_mw = 500.
        args.min_length = 7
        args.max_length = 25
        args.threads = 1
        args.max_variants_per_node = 9
        args.additional_variants_per_misc = 2
        args.min_nodes_to_collapse = 30
        args.naa_to_collapse = 5
        args.noncanonical_transcripts = False
        args.invalid_protein_as_noncoding = False
        args.debug_level = 1
        call_variant_peptide(args=args)

    def brute_force(self):
        """ call the brute force variant peptide caller """
        args = argparse.Namespace()
        args.input_gvf = [self.record.var_gvf_file]
        if self.record.circ_gvf_file.exists():
            args.input_gvf.append(self.record.circ_gvf_file)
        args.reference_dir = self.config.ref_dir
        args.force = True
        args.variant_ids = []
        args.max_adjacent_as_mnv = 2
        args.selenocysteine_termination = True
        args.w2f_reassignment = True
        args.cleavage_rule = self.config.cleavage_rule
        args.cleavage_exception = None
        args.miscleavage = self.config.miscleavage
        args.min_mw = self.config.min_mw
        args.min_length = self.config.min_length
        args.max_length = self.config.max_length

        with open(self.record.brute_force_fasta, 'wt') as handle:
            with redirect_stdout(handle):
                brute_force.main(args)

    def assert_equal(self) -> str:
        """ Assert that the callVariant results and bruteForce results equal """
        with open(self.record.call_variant_fasta, 'rt') as handle:
            variant_seqs = set()
            for seq in SeqIO.parse(handle, 'fasta'):
                variant_seqs.add(str(seq.seq))
        with open(self.record.brute_force_fasta, 'rt') as handle:
            brute_force_seqs = set()
            for line in handle:
                brute_force_seqs.add(line.rstrip())

        if variant_seqs != brute_force_seqs:
            variant_only = variant_seqs - brute_force_seqs
            brute_force_only = brute_force_seqs - variant_seqs
            if variant_only:
                with open(self.record.call_variant_only, 'wt') as handle:
                    for seq in variant_only:
                        handle.write(seq + '\n')
            if brute_force_only:
                with open(self.record.brute_force_only, 'wt') as handle:
                    for seq in brute_force_only:
                        handle.write(seq + '\n')
            return str(FuzzRecordStatus.failed)
        return str(FuzzRecordStatus.succeeded)

class FuzzTestConfig():
    """ Fuzz test config """
    def __init__(self, tx_id:str, n_iter:int, max_size:int, max_variants:int,
            min_variants:int, exonic_only:bool, snv_frac:float, snv_only_frac:float,
            fusion_frac:float, circ_rna_frac:float, ci_ratio:float, alt_splice_frac:bool,
            keep_succeeded:bool, save_graph:bool, cleavage_rule:str, cleavage_exception:str,
            miscleavage:int, min_mw:int, min_length:int, max_length:int, temp_dir:Path,
            output_dir:Path, ref_dir:Path, fuzz_start:datetime=None, fuzz_end:datetime=None,
            seed:int=None, nthreads:int=1):
        """ constructor """
        self.tx_id = tx_id
        self.n_iter = n_iter
        self.max_size = max_size
        self.max_variants = max_variants
        self.min_vairants = min_variants
        self.exonic_only = exonic_only
        self.snv_frac = snv_frac
        self.snv_only_frac = snv_only_frac
        self.fusion_frac = fusion_frac
        self.circ_rna_frac = circ_rna_frac
        self.ci_ratio = ci_ratio
        self.alt_splice_frac = alt_splice_frac
        self.keep_succeeded = keep_succeeded
        self.save_graph = save_graph
        self.cleavage_rule = cleavage_rule
        self.cleavage_exception = cleavage_exception
        self.miscleavage = miscleavage
        self.min_mw = min_mw
        self.min_length = min_length
        self.max_length = max_length
        self.temp_dir = temp_dir
        self.output_dir = output_dir
        self.ref_dir = ref_dir
        self.fuzz_start = fuzz_start
        self.fuzz_end = fuzz_end
        self.seed = seed
        self.nthreads = nthreads

    def to_dict(self):
        """ Convert to a dict """
        mapper = {}
        for k,v in self.__dict__.items():
            if isinstance(v, Path):
                v = str(v)
            elif isinstance(v, datetime):
                continue
            mapper[k] = v
        return mapper

    @property
    def path_genome_fasta(self) -> Path:
        """ Genome FASTA path """
        return self.ref_dir/'genome.fasta'

    @property
    def path_annotation_gtf(self) -> Path:
        """ Annotation FASTA path """
        return self.ref_dir/'annotation.gtf'

    @property
    def path_proteome_fasta(self) -> Path:
        """ Proteome FASTA path """
        return self.ref_dir/'proteome.fasta'

    @property
    def path_log_file(self) -> Path:
        """ Log file path """
        start_time = self.fuzz_start.strftime('%y%m%d-%H%M%S')
        return self.output_dir/f"fuzz-{start_time}.log"

class Fuzzer():
    """ Main class for fuzzing """
    def __init__(self, config:FuzzTestConfig, records:List[FuzzRecord]=None):
        """ constructor """
        self.config = config
        self.config.temp_dir.mkdir(exist_ok=True)
        self.config.output_dir.mkdir(exist_ok=True)
        self.records = records or []
        self.handle = None

    def __enter__(self):
        """ enter """
        self.set_fuzz_start_time()
        self.print_prolog()
        self.handle = self.config.path_log_file.open('wt')
        self.handle.write(FuzzRecord.tsv_header() + '\n')
        self.handle.flush()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ exit """
        self.handle.close()
        self.set_fuzz_end_time()
        self.print_epilog()

    def set_fuzz_start_time(self):
        """ set fuzz start time """
        self.config.fuzz_start = datetime.now()

    def set_fuzz_end_time(self):
        """ set fuzz end time """
        self.config.fuzz_end = datetime.now()

    def print_prolog(self):
        """ print prolog """
        fuzz_start = self.config.fuzz_start.strftime('%Y-%m-%d %H:%M:%S')
        plain_logger(f"moPepGen fuzz testing started at {fuzz_start}")

    def print_epilog(self):
        """ print epilog """
        fuzz_end = self.config.fuzz_end.strftime('%Y-%m-%d %H:%M:%S')
        plain_logger(f"moPepGen fuzz testing ended at {fuzz_end}")
        n_succeeded = 0
        n_failed = 0
        for x in self.records:
            if x.is_succeeded():
                n_succeeded += 1
            else:
                n_failed += 1
        plain_logger(f"{n_succeeded} cases passed.")
        plain_logger(f"{n_failed} cases failed.")

    def fuzz(self):
        """ fuzz """
        pool = ParallelPool(ncpus=self.config.nthreads)
        cases:List[FuzzTestCase] = []
        for i in range(self.config.n_iter):
            config = copy.deepcopy(self.config)
            case = FuzzTestCase(config)
            cases.append(case)
            if len(cases) >= config.nthreads or i == self.config.n_iter - 1:
                res:List[FuzzTestCase]
                if self.config.nthreads == 1:
                    res = map(_run_fuzz, cases)
                else:
                    res = pool.map(_run_fuzz, cases)
                for case in res:
                    self.handle.write(case.record.to_tsv() + '\n')
                    self.handle.flush()
                    self.records.append(case.record)
                cases = []

def _run_fuzz(case:FuzzTestCase) -> FuzzTestCase:
    """ Run fuzz test """
    case.run()
    return case

def main(args:argparse.Namespace):
    """ Main entry point for fuzz test """
    config = FuzzTestConfig(
        tx_id=args.tx_id,
        n_iter=args.n_iter,
        max_size=args.max_size,
        max_variants=args.max_variants,
        min_variants=args.min_variants,
        exonic_only=args.exonic_only,
        snv_frac=args.snv_frac,
        snv_only_frac=args.snv_only_frac,
        fusion_frac=args.fusion_frac,
        alt_splice_frac=args.alt_splice_frac,
        circ_rna_frac=args.circ_rna_frac,
        ci_ratio=args.ci_ratio,
        cleavage_rule=args.cleavage_rule,
        cleavage_exception=args.cleavage_exception,
        miscleavage=args.miscleavage,
        min_mw=args.min_mw,
        min_length=args.min_length,
        max_length=args.max_length,
        temp_dir=args.temp_dir,
        output_dir=args.output_dir,
        ref_dir=args.reference_dir,
        keep_succeeded=args.keep_succeeded,
        save_graph=args.save_graph,
        seed=args.seed,
        nthreads=args.nthreads
    )
    with Fuzzer(config) as fuzzer:
        fuzzer.fuzz()



""" In disk class for VariantRecordPool """
from __future__ import annotations
from typing import IO, Iterable, List, TYPE_CHECKING, Union
from moPepGen import circ
from . import io


# To avoid circular import
if TYPE_CHECKING:
    from moPepGen.circ.CircRNA import CircRNAModel
    from moPepGen.seqvar import VariantRecord


class GVFPointer():
    """ File pointer """
    def __init__(self, handle:IO, key:str, start:int=None, end:int=None,
            is_circ_rna:bool=False):
        """ constructor """
        self.handle = handle
        self.key = key
        self.start = start
        self.end = end
        self.is_circ_rna = is_circ_rna

    def __len__(self) -> int:
        """ length """
        return self.end - self.start

    def to_line(self) -> str:
        """ to line """
        return f"{self.key}\t{int(self.start)}\t{str(len(self))}"

    def __iter__(self) -> Iterable[Union[VariantRecord, CircRNAModel]]:
        """ Iterate variant records from handle """
        cur = self.handle.tell()
        offset = self.start - cur
        self.handle.seek(offset, 1)
        buffer:bytes = self.handle.read(len(self))
        buffer = buffer.decode('utf-8')
        for line in buffer.rstrip().split('\n'):
            if self.is_circ_rna:
                yield circ.io.line_to_circ_model(line)
            else:
                yield io.line_to_variant_record(line)

    def load(self) -> List[Union[VariantRecord, CircRNAModel]]:
        """ Load variant records from handle """
        return list(iter(self))

    @classmethod
    def parse(cls, index_handle:IO, gvf_handle:IO, is_circ_rna) -> Iterable[GVFPointer]:
        """ parse """
        line:str
        for line in index_handle:
            if line.startswith('#'):
                continue
            key, start, length = line.rstrip().split('\t')
            start = int(start)
            end = start + int(length)
            yield cls(handle=gvf_handle, key=key, start=start, end=end,
                is_circ_rna=is_circ_rna)

def iterate_pointer(handle, is_circ_rna:bool) -> Iterable[GVFPointer]:
    """ Generate GVF file poitners """
    cur_key = None
    line_end = 0
    pointer = None
    for line in handle:
        line:bytes
        line_start = line_end
        line_end += len(line)
        line = line.decode('utf-8')

        if line.startswith('#'):
            continue

        if is_circ_rna:
            record = circ.io.line_to_circ_model(line)
        else:
            record = io.line_to_variant_record(line)

        key = record.transcript_id

        if cur_key != key:
            if not cur_key is None:
                yield pointer
            cur_key = key
            pointer = GVFPointer(
                handle=handle, key=cur_key, start=line_start, end=line_end,
                is_circ_rna=is_circ_rna
            )
        else:
            pointer.end = line_end

    if pointer is not None:
        yield pointer



""" Module for CircularVariantGraph, a directed cyclic graph, for circRNA etc.
"""
from __future__ import annotations
import copy
from typing import Dict, Union, List, TYPE_CHECKING
from moPepGen.SeqFeature import FeatureLocation
from moPepGen import seqvar
from .ThreeFrameTVG import ThreeFrameTVG
from .TVGNode import TVGNode


if TYPE_CHECKING:
    from moPepGen.circ import CircRNAModel
    from moPepGen.dna import DNASeqRecordWithCoordinates
    from .SubgraphTree import SubgraphTree
    from moPepGen.params import CleavageParams
    from moPepGen.seqvar import VariantRecordWithCoordinate, VariantRecord

class ThreeFrameCVG(ThreeFrameTVG):
    """ Defines a directed cyclic graph for circular nucleotide molecules such
    ass circRNA and the variants associated with it.

    Attributes:
        seq (DNASeqRecordWithCoordinates): The original sequence of the
                circRNA (reference).
        transcript_id (str): The transcript ID that the circRNA is associated
            with
        attrs (dict): additional attributes
    """
    def __init__(self, seq:Union[DNASeqRecordWithCoordinates,None],
            _id:str, root:TVGNode=None, reading_frames:List[TVGNode]=None,
            cds_start_nf:bool=False, has_known_orf:bool=False,
            circ_record:CircRNAModel=None, attrs:dict=None,
            coordinate_feature_type:str=None, coordinate_feature_id:str=None,
            subgraphs:SubgraphTree=None, hypermutated_region_warned:bool=False,
            cleavage_params:CleavageParams=None,
            max_adjacent_as_mnv:int=2):
        """ Construct a CircularVariantGraph

        Args:
            seq (DNASeqRecordWithCoordinates): The original sequence of the
                    transcript (reference).
            transcript_id (str): The transcript ID that the circRNA is
                associated with.
            circ (CircRNAodel): The circRNA model.
            attrs (dict): additional attributes

        """
        self.seq = seq
        self.id = _id
        if self.seq and not self.seq.locations:
            self.add_default_sequence_locations()
        self.attrs = attrs
        self.circ = circ_record
        location = FeatureLocation(
            seqname=self.id,
            start=seq.locations[0].ref.start,
            end=seq.locations[0].ref.end
        )
        circ_variant = seqvar.VariantRecord(
            location=location,
            ref=seq.seq[0],
            alt='<circRNA>',
            _type='circRNA',
            _id=_id
        )
        super().__init__(
            seq=seq, _id=_id, root=root, reading_frames=reading_frames,
            cds_start_nf=cds_start_nf, has_known_orf=has_known_orf,
            global_variant=circ_variant, subgraphs=subgraphs,
            cleavage_params=cleavage_params, max_adjacent_as_mnv=max_adjacent_as_mnv,
            coordinate_feature_type=coordinate_feature_type,
            coordinate_feature_id=coordinate_feature_id,
            hypermutated_region_warned=hypermutated_region_warned
        )

    def get_circ_variant_with_coordinate(self) -> VariantRecordWithCoordinate:
        """ Add a variant record to the frameshifting of the root node. This
        will treat all peptides as variant peptide in the later steps. """
        location = FeatureLocation(
            seqname=self.circ.gene_id,
            start=0,
            end=len(self.seq)
        )
        return seqvar.VariantRecordWithCoordinate(self.global_variant, location)

    def init_three_frames(self, truncate_head:bool=False):
        """ Initiate the three reading-frame graph. """
        super().init_three_frames(truncate_head=truncate_head)
        var = self.get_circ_variant_with_coordinate()
        for i,root in enumerate(self.reading_frames):
            if len(root.out_edges) > 1:
                raise ValueError('Initiating CVG should not contain any variant')
            node = list(root.out_edges)[0].out_node
            var_i = copy.deepcopy(var)
            var_i.location.reading_frame_index = i
            node.variants.append(var_i)
            self.add_edge(node, root, 'reference')

    def create_variant_circ_graph(self, variants:List[VariantRecord]):
        """ Apply a list of variants to the graph. Variants not in the
        range are ignored. Variants at the first nucleotide of each fragment
        of the sequence are also ignored, because it causes the exon splice
        site to be changed.
        """
        filtered_variants = []
        exclude_type = ['Insertion', 'Deletion', 'Substitution', 'Fusion', 'circRNA']
        for variant in variants:
            if variant.type == exclude_type:
                continue

            circ_start = self.circ.fragments[0].location.start
            if circ_start <= variant.location.start < circ_start + 3:
                continue

            for fragment in self.circ.fragments:
                location = fragment.location
                if location.start <= variant.location.start < location.end:
                    filtered_variants.append(variant)
                    break
        super().create_variant_graph(filtered_variants, None, None, None)

    def get_tail_nodes(self) -> Dict[TVGNode, TVGNode]:
        """ The the last node of each reading frame. """
        tail_nodes:List[TVGNode] = [None, None, None]
        for i, root in enumerate(self.reading_frames):
            if len(root.in_edges) > 2:
                raise ValueError('CVG should not have any stop altering mutation')
            in_edge = None
            for in_edge in root.in_edges:
                if in_edge.in_node is not self.root:
                    break
            if not in_edge:
                raise ValueError('The TVG is not cyclic')
            tail_nodes[i] = in_edge.in_node
            self.remove_edge(in_edge)
        return tail_nodes

    def extend_loop(self):
        """ Extend the three frame circular graph into linear by repeating the
        graph of each reading frame for 3 additional copies. """
        n_loops = 3

        subgraphs:List[ThreeFrameCVG] = []
        cur = self
        for _ in range(n_loops):
            sub = cur.create_subgraph_copy()
            subgraphs.append(sub)
            cur = sub

        cur = self
        cur_tail_nodes = self.get_tail_nodes()
        cur = self
        for sub in subgraphs:
            sub_tail_nodes = sub.get_tail_nodes()
            self.subgraphs.add_subgraph(
                child_id=sub.id, parent_id=cur.id, level=sub.root.level,
                start=self.seq.locations[-1].ref.end, end=self.seq.locations[-1].ref.end,
                variant=self.global_variant, feature_type='gene', feature_id=self.circ.gene_id
            )
            for edge in sub.root.out_edges:
                root = edge.out_node
                if len(root.out_edges) > 1:
                    raise ValueError('CVG should not have any start altering mutation')
                head = list(root.out_edges)[0].out_node
                end_node = cur_tail_nodes[head.reading_frame_index]
                for _edge in copy.copy(head.in_edges):
                    cur.remove_edge(_edge)
                cur.add_edge(end_node, head, 'reference')
            cur = sub
            cur_tail_nodes = sub_tail_nodes

    def truncate_three_frames(self):
        """ For each of the reading frames, tuncate the leading nodes for 1
        nucleotide. """
        for i, root in enumerate(self.reading_frames):
            if i == 0:
                continue
            if len(root.out_edges) > 1:
                raise ValueError('CVG should not contain any start altering mutaiton')
            head = list(root.out_edges)[0].out_node
            head.truncate_left(i)

    def create_subgraph_copy(self) -> ThreeFrameCVG:
        """ Create a subgraph by copying the current graph. The subgraph created
        will be attached to the end of this current graph. """
        subgraph_id = self.subgraphs.generate_subgraph_id()
        seq = copy.deepcopy(self.seq)
        for loc in seq.locations:
            loc.ref.seqname = subgraph_id
        root = self.root.deepcopy(subgraph_id, 1)
        reading_frames:Dict[TVGNode] = [None, None, None]
        for node in root.get_out_nodes():
            reading_frames[node.reading_frame_index] = node
        return ThreeFrameCVG(
            seq=seq, _id=subgraph_id, root=root, reading_frames=reading_frames,
            cds_start_nf=self.cds_start_nf, has_known_orf=self.has_known_orf,
            circ_record=self.circ, attrs=copy.copy(self.attrs),
            subgraphs=None, cleavage_params=self.cleavage_params
        )



""" IO """
from __future__ import annotations
from typing import Iterable, List, IO, TYPE_CHECKING
from moPepGen import GVF_HEADER
from moPepGen.SeqFeature import FeatureLocation, SeqFeature
from .CircRNA import CircRNAModel


if TYPE_CHECKING:
    from moPepGen.seqvar import GVFMetadata

def parse(handle:IO) -> Iterable[CircRNAModel]:
    """ Parse a circRNA TSV file and returns an iterable of CircRNAModel
    object """
    for line in handle:
        if line.startswith('#'):
            continue
        yield line_to_circ_model(line)

def line_to_circ_model(line) -> CircRNAModel:
    """ Convert GVF line to CircRNAModel """
    fields = line.rstrip().split('\t')

    gene_id = fields[0]
    start = int(fields[1])
    circ_id = fields[2]

    attrs = {}
    for attr in fields[7].split(';'):
        key,val = attr.split('=')
        if key in ['OFFSET', 'LENGTH']:
            val = [int(x) for x in val.split(',')]
        elif key == 'INTRON':
            val = [] if val == '' else [int(x) for x in val.split(',')]
        attrs[key] = val

    offsets = attrs['OFFSET']
    lengths = attrs['LENGTH']
    introns = attrs['INTRON']
    tx_id = attrs['TRANSCRIPT_ID']
    gene_name = attrs['GENE_SYMBOL']
    genomic_location = attrs.get('GENOMIC_LOCATION', '')

    fragments:List[SeqFeature] = []
    for j, (position, length) in enumerate(zip(offsets, lengths)):
        start_j = start + position
        end_j = start_j + length
        location = FeatureLocation(seqname=gene_id, start=start_j, end=end_j)
        frag = SeqFeature(
            chrom=gene_id, location=location, attributes={},
            type='intron' if j+1 in introns else 'exon'
        )
        fragments.append(frag)

    return CircRNAModel(tx_id, fragments, introns, circ_id, gene_id, gene_name,
            genomic_location)

def write(records:Iterable[CircRNAModel], metadata:GVFMetadata, handle:IO):
    """ Write circRNA records to file. """
    metadata.add_info('circRNA')
    for line in metadata.to_strings():
        handle.write(line + '\n')
    handle.write('#' + '\t'.join(GVF_HEADER) + '\n')
    for record in records:
        line = record.to_string()
        handle.write(line + '\n')



""" Module for common functions for the util module """
from pathlib import Path
from typing import Tuple
from moPepGen.gtf import GenomicAnnotation
from moPepGen.dna import DNASeqDict
from moPepGen.aa import AminoAcidSeqDict


def load_references(path_anno:Path, path_genome:Path, path_proteome:Path
        ) -> Tuple[GenomicAnnotation, DNASeqDict, AminoAcidSeqDict]:
    """ Load reference files """
    anno = GenomicAnnotation()
    anno.dump_gtf(path_anno)

    genome = DNASeqDict()
    genome.dump_fasta(path_genome)

    proteome = AminoAcidSeqDict()
    proteome.dump_fasta(path_proteome)

    anno.check_protein_coding(proteome, invalid_protein_as_noncoding=True)

    return anno, genome, proteome



""" Module for Transcript Annotation model """
from __future__ import annotations
from typing import List, TYPE_CHECKING
from moPepGen.SeqFeature import FeatureLocation, MatchedLocation
from moPepGen.gtf.GTFSeqFeature import GTFSeqFeature
from moPepGen import dna, ERROR_INDEX_IN_INTRON


if TYPE_CHECKING:
    from Bio.Seq import Seq

GTF_FEATURE_TYPES = {
    'transcript': 'transcript',
    'cds': 'cds',
    'exon': 'exon',
    'start_codon': 'start_codon',
    'stop_codon': 'stop_codon',
    'utr': 'utr',
    'selenocysteine': 'selenocysteine',
    'five_prime_utr': 'five_utr',
    'three_prime_utr': 'three_utr'
}

class TranscriptAnnotationModel():
    """ A TranscriptAnnotationModel holds all the annotations associated with
    the same transcript from a GTF file.

    Attributes:
        transcript (GTFRecord): The GTF record of the transcript
        cds (List[GTFRecord]): The GTF records of all the coding sequences of
            the transcript.
        exon (List[GTFRecord]): The GTF records of all the exons of the
            transcript.
        start_codon (List[GTFRecord]): The GTF records of all the start codons
            of the transcript.
        stop_codon (List[GTFRecord]): The GTF records of all the stop codons
            of the transcript.
    """
    def __init__(self, transcript:GTFSeqFeature=None,
            cds:List[GTFSeqFeature]=None, exon:List[GTFSeqFeature]=None,
            start_codon:List[GTFSeqFeature]=None,
            stop_codon:List[GTFSeqFeature]=None,
            utr:List[GTFSeqFeature]=None,
            five_utr:List[GTFSeqFeature]=None,
            three_utr:List[GTFSeqFeature]=None,
            selenocysteine:List[GTFSeqFeature]=None,
            is_protein_coding:bool=None,
            _seq:dna.DNASeqRecordWithCoordinates=None,
            transcript_id:str=None,
            gene_id:str=None,
            protein_id:str=None,
            gene_name:str=None,
            gene_type:str=None):
        """ Construct a TranscriptAnnotationmodel """
        self.transcript = transcript
        self.cds = cds or []
        self.exon = exon or []
        self.start_codon = start_codon or []
        self.stop_codon = stop_codon or []
        self.utr = utr or []
        self.five_utr = five_utr or []
        self.three_utr = three_utr or []
        self.selenocysteine = selenocysteine or []
        self.is_protein_coding = is_protein_coding
        self._seq = _seq
        self.transcript_id = transcript_id
        self.gene_id = gene_id
        self.protein_id = protein_id
        self.gene_name = gene_name
        self.gene_type = gene_type

    def add_record(self, _type:str, record: GTFSeqFeature):
        """ Add a GTFRecrod into a TranscriptAnnotationModel. If the biotype
        is cds, exon, start_codon, or stop_codon, it is inserted in order.

        Args:
            type (str): Type of annotation to add. Must be from transcript,
                cds, exon, start_codon, or stop_codon.
            record (GTFRecord): The GTF record to be added.
        """
        if _type not in GTF_FEATURE_TYPES:
            raise ValueError(f'Type must be from {list(GTF_FEATURE_TYPES.keys())}')
        attr = GTF_FEATURE_TYPES[_type]

        for key in ['transcript_id', 'gene_id', 'protein_id', 'gene_name', 'gene_type']:
            if hasattr(record, key):
                if self.__getattribute__(key) is None:
                    val = record.__getattribute__(key)
                    if val is not None and val != '':
                        self.__setattr__(key, record.__getattribute__(key))
                else:
                    record.__setattr__(key, self.__getattribute__(key))

        if _type == 'transcript':
            self.transcript = record
            if 'is_protein_coding' in record.attributes:
                is_protein_coding = record.attributes.pop('is_protein_coding')
                self.is_protein_coding = is_protein_coding == 'true'
        else:
            if self.__getattribute__(attr) is None:
                self.__setattr__(attr, [])
            self.__getattribute__(attr).append(record)

    def split_utr(self) -> None:
        """ Infer 3'UTR or 5'UTR. CDS must be already sorted. """
        if not self.utr:
            return
        if not self.cds:
            raise ValueError(
                "UTR found but not CDS for transcript" + \
                    self.transcript.transcript_id
            )
        strand = self.transcript.strand
        for utr in self.utr:
            if strand == 1 and utr < self.cds[0] \
                    or strand == -1 and utr > self.cds[-1]:
                self.five_utr.append(utr)
            else:
                self.three_utr.append(utr)

    def remove_cached_seq(self):
        """ Remove the cased sequence """
        self._seq = None

    def sort_records(self):
        """ sort records """
        self.cds.sort()
        self.exon.sort()
        self.start_codon.sort()
        self.stop_codon.sort()
        self.split_utr()
        self.utr.sort()
        self.three_utr.sort()
        self.five_utr.sort()
        self.selenocysteine.sort()

    def get_cds_start_index(self) -> int:
        """ Returns the CDS start index of the transcript """
        cds_start = 0
        if self.transcript.strand == 1:
            cds_frame = self.cds[0].frame
            for exon in self.exon:
                if self.cds[0].location.start not in exon:
                    cds_start += len(exon)
                    continue
                cds_start += (self.cds[0].location.start - exon.location.start)
                break
        elif self.transcript.strand == -1:
            cds_frame = self.cds[-1].frame or 0
            for exon in reversed(self.exon):
                if self.cds[-1].location.end != exon.location.end and \
                        self.cds[-1].location.end not in exon:
                    cds_start += len(exon)
                    continue
                cds_start += (exon.location.end - self.cds[-1].location.end)
                break
        else:
            raise ValueError('Strand must not be unknown.')
        return cds_start + cds_frame

    def get_cds_end_index(self, seq:Seq, start:int) -> int:
        """ Returns the CDS stop index of the transcript. """
        if self.three_utr:
            if self.transcript.strand == 1:
                end = self.get_transcript_index(self.three_utr[0].location.start)
            else:
                end = self.get_transcript_index(self.three_utr[-1].location.end - 1)
            return end - (end - start) % 3
        return len(seq) - (len(seq) - start) % 3

    def get_transcript_sequence(self, chrom:dna.DNASeqRecord,
            cache:bool=False) -> dna.DNASeqRecordWithCoordinates:
        """ Returns the transcript sequence. The is done by concating all the
        exon sequences. If the gene is on the negatice strand, the reverse
        complement is returned.

        Args:
            chrom (DNASeqRecord): The chromosome sequence that the transcript
                is located.
        """
        if self._seq:
            return self._seq
        if len(self.exon) == 0:
            raise ValueError("Transcript model has no exon")
        seq = None
        for location in [it.location for it in self.exon]:
            new_seq = chrom.seq[location.start:location.end]
            if seq is None:
                seq = new_seq
            else:
                seq = seq + new_seq

        if self.transcript.strand == -1:
            seq = seq.reverse_complement()

        if self.cds:
            cds_start = self.get_cds_start_index()
            cds_end = self.get_cds_end_index(seq, cds_start)
            orf = FeatureLocation(start=cds_start, end=cds_end)
        else:
            orf = None

        selenocystein = []
        for sec in self.selenocysteine:
            if sec.strand == 1:
                sec_start = self.get_transcript_index(sec.location.start)
                sec_end = self.get_transcript_index(sec.location.end - 1) + 1
            else:
                sec_start = self.get_transcript_index(sec.location.end - 1)
                sec_end = self.get_transcript_index(sec.location.start) + 1
            selenocystein.append(FeatureLocation(start=sec_start, end=sec_end))
        selenocystein.sort()

        location = MatchedLocation(
            query=FeatureLocation(start=0, end=len(seq)),
            ref=FeatureLocation(
                seqname=self.transcript.transcript_id,
                start=0,
                end=len(seq)
            )
        )
        transcript = dna.DNASeqRecordWithCoordinates(
            seq=seq,
            orf=orf,
            selenocysteine=selenocystein,
            locations=[location]
        )
        transcript.id = self.transcript.transcript_id
        transcript.name = self.transcript.transcript_id
        transcript.description = self.transcript.transcript_id +\
            '|' + self.transcript.gene_id
        if 'protein_id' in self.transcript.attributes:
            transcript.description += '|' + \
                self.transcript.protein_id
        if cache:
            self._seq = transcript
        return transcript

    def get_cdna_sequence(self, chrom:dna.DNASeqRecord
            ) -> dna.DNASeqRecordWithCoordinates:
        """ Get the cDNA sequence.

        Args:
            chrom (DNASeqRecord): The chromosome sequence.

        Returns:
            The cDNA sequence (DNASeqRecordWithCoordinates)
        """
        if len(self.cds) == 0:
            raise ValueError('Transcript model has no cds')
        seq = None
        locations = [it.location for it in self.cds]
        for location in locations:
            new_seq = chrom.seq[location.start:location.end]
            if seq is None:
                seq = new_seq
            else:
                seq = seq + new_seq

        cds_start = self.get_cds_start_index()

        if self.transcript.strand == -1:
            seq = seq.reverse_complement()

        location = MatchedLocation(
            query=FeatureLocation(start=0, end=len(seq)),
            ref=FeatureLocation(
                seqname=self.transcript.transcript_id,
                start=cds_start,
                end=cds_start + len(seq)
            )
        )
        cdna = dna.DNASeqRecordWithCoordinates(
            seq=seq,
            locations=[location]
        )
        cdna.id = self.transcript.transcript_id
        cdna.name = self.transcript.transcript_id
        cdna.description = self.transcript.transcript_id + '|' \
            + self.transcript.gene_id + '|' \
            + self.transcript.protein_id
        return cdna

    def get_transcript_index(self, genomic_index:int) -> int:
        """ Get the transcript index from a genomic index. """
        if genomic_index < self.exon[0].location.start \
                or genomic_index >= self.exon[-1].location.end:
            raise ValueError(
                r'The genomic index isn\'t in the range of this transcript'
            )
        if self.transcript.strand == 1:
            index = 0
            for exon in self.exon:
                if exon.location.end < genomic_index:
                    index += exon.location.end - exon.location.start
                elif exon.location.end == genomic_index:
                    raise ValueError(ERROR_INDEX_IN_INTRON)
                elif exon.location.start <= genomic_index:
                    index += genomic_index - exon.location.start
                    break
                else:
                    raise ValueError(ERROR_INDEX_IN_INTRON)
        elif self.transcript.strand == -1:
            index = -1
            for exon in reversed(self.exon):
                if exon.location.start >= genomic_index:
                    index += exon.location.end - exon.location.start
                    if exon.location.start == genomic_index:
                        break
                elif exon.location.end > genomic_index:
                    index += exon.location.end - genomic_index
                    break
                else:
                    raise ValueError(ERROR_INDEX_IN_INTRON)
        return index

    def get_transcript_start_genomic_coordinate(self) -> int:
        """ Get the genomic coordinate of the start point of the transcript """
        if self.transcript.location.strand == 1:
            return int(self.exon[0].location.start)
        return int(self.exon[-1].location.end)

    def is_cds_start_nf(self) -> bool:
        """ Returns if the transcript has the tag of cds_start_NF """
        return 'tag' in self.transcript.attributes and \
            'cds_start_NF' in self.transcript.attributes['tag']

    def is_mrna_end_nf(self) -> bool:
        """ Returns if the transcript has the tag of mrna_end_NF """
        return 'tag' in self.transcript.attributes and \
            'mRNA_end_NF' in self.transcript.attributes['tag']

    def transcript_len(self) -> int:
        """ Get the transcript length minus introns """
        length = 0
        for exon in self.exon:
            length += exon.location.end - exon.location.start
        return length

    def is_exonic(self, pos:int) -> bool:
        """ Checks if the given genomic position is contained by any exon of
        the current transcript model.

        Args:
            pos (int): A genomic position
        """
        for exon in self.exon:
            if exon.location.start <= pos < exon.location.end:
                return True
        return False

    def get_next_exon(self, pos:int) -> GTFSeqFeature:
        """ Get the next exon downstream to the position. The position given
        must be genomic position """
        if self.transcript.strand == 1:
            for exon in self.exon:
                if exon.location.start > pos:
                    return exon
        else:
            for exon in reversed(self.transcript.exon):
                if exon.location.start <= pos:
                    return exon
        return None

    def get_upstream_exon_end(self, pos:int) -> int:
        """ Find the upstream exon end """

        if self.transcript.strand == 1:
            for exon in self.exon:
                if exon.location.end > pos:
                    break
                ind = exon.location.end - 1
        else:
            for exon in reversed(self.exon):
                if exon.location.start < pos:
                    break
                ind = exon.location.start
        if ind == -1:
            raise ValueError("Could not find the upstream exon end.")
        return ind

    def get_downstream_exon_start(self, pos:int) -> int:
        """ Find the downstream exon end """
        ind = -1
        if self.transcript.strand == 1:
            for exon in self.exon:
                if exon.location.start >= pos:
                    ind = exon.location.start
                    break
        else:
            for exon in reversed(self.exon):
                if exon.location.end - 1 <= pos:
                    ind = exon.location.end - 1
                    break
        if ind == -1:
            raise ValueError("Could not find the upstream exon end.")
        return ind

    def get_exon_with_start(self, start:int, offset:int=0) -> int:
        """ Get the exon with a given START position. """
        for i, exon in enumerate(self.exon[offset:]):
            if exon.location.start == start:
                return i + offset
        return -1

    def get_exon_with_end(self, end:int, offset:int=0) -> int:
        """ Get the exon with a given end position. """
        for i, exon in enumerate(self.exon[offset:]):
            if exon.location.end == end:
                return i + offset
        return -1

    def get_exon_overlaps(self, loc:FeatureLocation, offset:int=0) -> int:
        """ Get exons that overlap with a given location """
        indices = []
        for i, exon in enumerate(self.exon[offset:]):
            if exon.location.overlaps(loc):
                indices.append(i + offset)
        return indices

    def get_exon_containing(self, pos:int) -> int:
        """ Get exon that contains a position """
        for i, exon in enumerate(self.exon):
            if pos in exon.location:
                return i
        return -1

    def get_exon_inner(self, loc:FeatureLocation, offset:int=0) -> int:
        """ get exon that are inner subset of a given location """
        indices = []
        for i, exon in enumerate(self.exon[offset:]):
            if loc.is_superset(exon.location):
                indices.append(i + offset)
        return indices

    def has_junction(self, junction:FeatureLocation) -> bool:
        """ Checks if has the given junction """
        for i, exon1 in enumerate(self.exon):
            if exon1 is self.exon[-1]:
                break
            exon2 = self.exon[i + 1]
            if exon2.location.start > junction.end:
                break
            if exon1.location.end == junction.start \
                    and exon2.location.start == junction.end:
                return True
        return False



""" Infer GTF source (e.g. GENCODE/ENSEMBL) """
from __future__ import annotations
from typing import Dict, TYPE_CHECKING


if TYPE_CHECKING:
    from moPepGen.gtf.GTFSeqFeature import GTFSeqFeature

class GTFSourceInferrer():
    """ Infer GTF source (e.g. GENOCDE/ENSEMBL) """
    def __init__(self):
        """ Constructor """
        self.max_iter = 100
        self.data:Dict[str,int] = {}
        self.count = 0
        self.source:str = None

    def infer(self, record:GTFSeqFeature) -> str:
        """ Infer the source of a GTF record """
        if self.count > self.max_iter:
            if not self.source:
                self.source = sorted(self.data.items(), key=lambda x:x[1])[-1][0]
            return self.source
        self.count += 1
        record.infer_annotation_source()
        source = record.source
        if source not in self.data:
            self.data[source] = 1
        else:
            self.data[source] += 1
        return source



""" module for peptide peptide labels """
from __future__ import annotations
from typing import Dict, Iterable, List, TYPE_CHECKING, Set
from moPepGen import err, seqvar, circ, constant, SPLIT_DATABASE_KEY_SEPARATER
from . import VariantPeptideIdentifier as pi

if TYPE_CHECKING:
    from .AminoAcidSeqRecord import AminoAcidSeqRecord
    from moPepGen.gtf import GenomicAnnotation


class VariantSourceSet(set):
    """ Variant source set. This is a class of ordered set.

    Example:
        >>> VariantSourceSet.set_levels({'A':0, 'B':1, 'C':2})
        >>> print(VariantSourceSet(['A']) < VariantSourceSet(['B']))
        True

        >>> print(VariantSourceSet(['A', 'B']) > VariantSourceSet(['B']))
        True
    """
    levels_map = {}
    levels = []

    def __init__(self, *args):
        """ constructor """
        for it in list(*args):
            self.validate(it)
        super().__init__(*args)

    def validate(self, it):
        """ Validate element """
        if it not in self.levels_map:
            raise ValueError(f'No defined level found for {it}')

    def add(self, element, group_map:Dict[str,str]=None):
        """ override the add method """
        if group_map is None:
            group_map = {}
        if element in group_map:
            element = group_map[element]
        self.validate(element)
        super().add(element)

    @classmethod
    def set_levels(cls, levels:Dict[str,int]):
        """ set levels. This is to make sure that all instances share the same
        order. """
        cls.levels_map = levels
        cls.levels = [y[0] for y in sorted(cls.levels_map.items(), key=lambda x:x[1])]

    @classmethod
    def reset_levels(cls):
        """ reset levels """
        cls.levels_map = {}
        cls.levels = []

    def __str__(self) -> str:
        """ str """
        sorted_list = [x for x in self.levels if x in self]
        return SPLIT_DATABASE_KEY_SEPARATER.join(sorted_list)

    def __gt__(self, other:VariantSourceSet) -> bool:
        """ greater than """
        if self == other:
            return False
        this = self.to_int()
        that = other.to_int()
        if len(this) > len(that):
            return True
        if len(this) < len(that):
            return False
        for i, j in zip(this, that):
            if i > j:
                return True
            if i < j:
                return False
        return False

    def __ge__(self, other:VariantSourceSet) -> bool:
        """ greater or equal to """
        return self == other or self > other

    def __lt__(self, other:VariantSourceSet) -> bool:
        """ less then """
        return not self >= other

    def __le__(self, other:VariantSourceSet) -> bool:
        """ less or equal to """
        return not self > other

    def to_int(self, sort=True) -> Iterable[int]:
        """ to int """
        try:
            source_int = {self.levels_map[frozenset(self)]}
        except KeyError:
            source_int = {self.levels_map[x] for x in self}
        if sort:
            source_int = list(source_int)
            source_int.sort()
        return source_int

def is_circ_rna(_id:str) -> bool:
    """ Check if the id is a circRNA """
    return _id.startswith('CIRC-') or _id.startswith('CI-')

class VariantPeptideInfo():
    """ Variant peptide label. This is a helper class in order to sort peptide
    labels easily.

    Args:
        transcript_id (str): If the variant is a circRNA, this is the circRNA
            ID.
    """
    def __init__(self, orignial_label:str, gene_ids:List[str],
            variant_labels:Dict[str,List[str]],
            variant_index:int, sources:VariantSourceSet=None):
        """ Constructor """
        self.original_label = orignial_label
        self.gene_ids = gene_ids
        self.variant_labels = variant_labels
        self.variant_index = variant_index
        self.sources = sources or VariantSourceSet()


    @staticmethod
    def from_variant_peptide_minimal(peptide:AminoAcidSeqRecord
            ) -> List[VariantPeptideInfo]:
        """ Create list of VariantPeptideInfo with minimal information. """
        info_list:List[VariantPeptideInfo] = []
        variant_ids = pi.parse_variant_peptide_id(peptide.description, set())
        for variant_id in variant_ids:
            if isinstance(variant_id, pi.NovelORFPeptideIdentifier):
                gene_ids = [variant_id.gene_id]
                var_ids = {}

            elif isinstance(variant_id, pi.CircRNAVariantPeptideIdentifier):
                gene_ids = None
                var_ids = {}

            elif isinstance(variant_id, pi.FusionVariantPeptideIdentifier):
                first_tx_id = variant_id.first_tx_id
                second_tx_id = variant_id.second_tx_id
                gene_ids = None
                var_ids = {
                    first_tx_id: variant_id.first_variants + [variant_id.fusion_id],
                    second_tx_id: variant_id.second_variants
                }

            elif isinstance(variant_id, pi.BaseVariantPeptideIdentifier):
                gene_ids = None
                var_ids = {}

            info = VariantPeptideInfo(str(variant_id), gene_ids, var_ids, variant_id.index)

            info_list.append(info)
        return info_list

    @staticmethod
    def from_variant_peptide(peptide:AminoAcidSeqRecord,
            tx2gene:Dict[str,str], coding_tx:Set[str],
            label_map:LabelSourceMapping=None,
            check_source:bool=True, group_map:Dict[str,str]=None
            ) -> List[VariantPeptideInfo]:
        """ Parse from a variant peptide record """
        if group_map is None:
            group_map = {}
        info_list = []
        variant_ids = pi.parse_variant_peptide_id(peptide.description, coding_tx)
        for variant_id in variant_ids:
            if isinstance(variant_id, pi.NovelORFPeptideIdentifier):
                tx_id = variant_id.transcript_id
                gene_ids = [variant_id.gene_id]
                var_ids = {variant_id.gene_id: variant_id.codon_reassigns}

            elif isinstance(variant_id, pi.CircRNAVariantPeptideIdentifier):
                circ_rna_id = variant_id.circ_rna_id
                tx_id = circ_rna_id.split('-', 2)[1]
                gene_ids = [tx2gene[tx_id]]
                var_ids = {gene_ids[0]: [circ_rna_id, *variant_id.variant_ids]}

            elif isinstance(variant_id, pi.FusionVariantPeptideIdentifier):
                first_tx_id = variant_id.first_tx_id
                first_gene_id = tx2gene[first_tx_id]
                second_tx_id = variant_id.second_tx_id
                second_gene_id = tx2gene[second_tx_id]
                gene_ids = [first_gene_id, second_gene_id]
                var_ids = {
                    first_gene_id: variant_id.first_variants
                        + [variant_id.fusion_id] \
                        + variant_id.peptide_variants,
                    second_gene_id: variant_id.second_variants
                }
                tx_id = first_tx_id

            elif isinstance(variant_id, pi.BaseVariantPeptideIdentifier):
                gene_id = tx2gene[variant_id.transcript_id]
                gene_ids = [gene_id]
                var_ids = {gene_id: variant_id.variant_ids}
                tx_id = variant_id.transcript_id

            info = VariantPeptideInfo(str(variant_id), gene_ids, var_ids, variant_id.index)

            if check_source:
                if tx_id not in coding_tx:
                    info.sources.add(constant.SOURCE_NOVEL_ORF, group_map=group_map)

                for gene_id, _ids in var_ids.items():
                    for var_id in _ids:
                        if var_id.split('-')[0] == constant.SEC_TERMINATION_TYPE:
                            info.sources.add(
                                constant.SOURCE_SEC_TERMINATION,
                                group_map=group_map
                            )
                        elif var_id.split('-')[0] in constant.CODON_REASSIGNMENTS_TYPES:
                            info.sources.add(
                                constant.SOURCE_CODON_REASSIGNMENT,
                                group_map=group_map
                            )
                        else:
                            source = label_map.get_source(gene_id, var_id)
                            info.sources.add(source, group_map=group_map)

            info_list.append(info)
        return info_list

    def is_fusion(self) -> bool:
        """ Check if this is a fusion """
        _id = pi.parse_variant_peptide_id(self.original_label, set())[0]
        return isinstance(_id, pi.FusionVariantPeptideIdentifier)

    def is_circ_rna(self) -> bool:
        """ Check if this is a circRNA """
        _id = pi.parse_variant_peptide_id(self.original_label, set())[0]
        return isinstance(_id, pi.CircRNAVariantPeptideIdentifier)

    def is_splice_altering(self) -> bool:
        """ Check if the variant paptide label is alternative splicing """
        _id = pi.parse_variant_peptide_id(self.original_label, set())[0]
        return isinstance(_id, pi.BaseVariantPeptideIdentifier) and \
            _id.is_alternative_splicing()

    @staticmethod
    def is_noncoding(tx_id:str, inclusion:List[str], exclusion:List[str]) -> bool:
        """ Check if a transcript is novel ORF """
        if inclusion:
            return tx_id in inclusion
        if exclusion:
            return tx_id not in exclusion
        raise ValueError('At least one of inclusion or exclusion must given')


    def all_noncoding(self, anno:GenomicAnnotation) -> bool:
        """ Check if all transcripts are novel ORF """
        tx_ids = self.get_transcript_ids()
        return all(not anno.transcripts[tx_id].is_protein_coding for tx_id in tx_ids)

    def any_noncoding(self, anno:GenomicAnnotation) -> bool:
        """ Check if all transcripts are novel ORF """
        tx_ids = self.get_transcript_ids()
        return any(not anno.transcripts[tx_id].is_protein_coding for tx_id in tx_ids)

    def all_coding(self, anno:GenomicAnnotation) -> bool:
        """ Check if all transcripts are coding """
        tx_ids = self.get_transcript_ids()
        return all(anno.transcripts[tx_id].is_protein_coding for tx_id in tx_ids)

    def any_coding(self, anno:GenomicAnnotation) -> bool:
        """ Check if all transcripts are coding """
        tx_ids = self.get_transcript_ids()
        return any(anno.transcripts[tx_id].is_protein_coding for tx_id in tx_ids)

    def get_transcript_ids(self) -> List[str]:
        """ get transcript IDs """
        variant_id = pi.parse_variant_peptide_id(self.original_label, set())[0]
        if isinstance(variant_id, pi.CircRNAVariantPeptideIdentifier):
            return [variant_id.circ_rna_id.split('-', 2)[1]]
        if isinstance(variant_id, pi.FusionVariantPeptideIdentifier):
            return [variant_id.first_tx_id, variant_id.second_tx_id]
        if isinstance(variant_id, pi.BaseVariantPeptideIdentifier):
            return [variant_id.transcript_id]
        if isinstance(variant_id, pi.NovelORFPeptideIdentifier):
            return [variant_id.transcript_id]
        raise ValueError('Variant ID unrecognized')

    def __str__(self) -> str:
        """ str """
        return self.original_label

    def __eq__(self, other:VariantPeptideInfo):
        """ equal to """
        if len(self.gene_ids) != len(other.gene_ids):
            return False
        if any(x != y for x,y in zip(self.gene_ids, other.gene_ids)):
            return False
        return set(self.variant_labels) == set(other.variant_labels) \
            and self.variant_index == other.variant_index \
            and self.sources == other.sources

    def __ne__(self, other:VariantPeptideInfo):
        """ not equal to """
        return not self == other

    def __gt__(self, other:VariantPeptideInfo):
        """" greater than """
        return self.sources > other.sources

    def __ge__(self, other:VariantPeptideInfo):
        """ greater than or equal to """
        return self > other or self == other

    def __lt__(self, other:VariantPeptideInfo):
        """ less than """
        return not self >=  other

    def __le__(self, other:VariantPeptideInfo):
        """ less than or equal to """
        return not self > other

class LabelSourceMapping():
    """ Helper class to handle label source mapping """
    def __init__(self, data:Dict[str,Dict[str,str]]=None):
        """ constructor"""
        self.data = data or {}

    def add_record(self, gene_id:str, label:str, source:str):
        """ add a record """
        if gene_id not in self.data:
            self.data[gene_id] = {}
            if label in self.data[gene_id]:
                raise ValueError(
                    f"Duplicated variable label found for {gene_id}, {label}"
                )
        self.data[gene_id][label] = source

    def add_variant(self, variant:seqvar.VariantRecord, source:str):
        """ add variant """
        gene_id = variant.location.seqname
        if gene_id not in self.data:
            self.data[gene_id] = {}
        if variant.id not in self.data[gene_id]:
            self.data[gene_id][variant.id] = source

    def add_circ_rna(self, record:circ.CircRNAModel, source:str):
        """ add circRNA record """
        gene_id = record.gene_id
        if gene_id not in self.data:
            self.data[gene_id] = {}
        if record.id not in self.data[gene_id]:
            self.data[gene_id][record.id] = source

    def get_source(self, gene_id:str, var_id:str) -> str:
        """ Get source """
        try:
            return self.data[gene_id][var_id]
        except KeyError as e:
            raise err.VariantSourceNotFoundError(gene_id, var_id) from e



""" Module for GTF """
from .GTFRecord import GTFRecord
from .TranscriptAnnotationModel import TranscriptAnnotationModel, GTF_FEATURE_TYPES
from .GenomicAnnotation import GenomicAnnotation
from .GeneAnnotationModel import GeneAnnotationModel
from .GenomicAnnotationOnDisk import GenomicAnnotationOnDisk



""" moPepGen """
from __future__ import annotations
from datetime import datetime
import hashlib
import itertools
from typing import Iterable, IO
import logging
from . import constant


__version__ = '1.4.1'

## Error messages
ERROR_INDEX_IN_INTRON = 'The genomic index seems to be in an intron'
ERROR_NO_TX_AVAILABLE = 'No transcripts available'
ERROR_VARIANT_NOT_IN_GENE_COORDINATE = 'Variant is not in gene coordinates.'
ERROR_REF_LENGTH_NOT_MATCH_WITH_LOCATION = \
    'Length of ref does not match with the location.'
# global variables

# Used in the moPepGen.aa.VariantPeptidePool module to serve as a delimiter of
# variant peptide seqname if it is found in multiple transcripts. E.g.:
# ENSG0001|SNV-10-T-A|1 ENSG0002|INDEL-20-CCCC-C|2
VARIANT_PEPTIDE_SOURCE_DELIMITER = ' '

# Used in moPepGen.aa.PeptidePoolSplitter to server as the separater in split
# database keys. E.g.:
# gSNP-gINDEL
SPLIT_DATABASE_KEY_SEPARATER = '-'

# GVF header
GVF_HEADER = ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO']

class _CaptureEq:
    """Object wrapper that remembers "other" for successful equality tests.
    Adopted from https://code.activestate.com/recipes/499299/
    """
    def __init__(self, obj):
        self.obj = obj
        self.match = obj

    def __eq__(self, other):
        result = (self.obj == other)
        if result:
            self.match = other
        return result

    def __getattr__(self, name):  # support hash() or anything else needed by __contains__
        return getattr(self.obj, name)

    def __hash__(self):
        return hash(self.obj)

def get_equivalent(container, item, default=None):
    '''Gets the specific container element matched by: "item in container".

    Adopted from https://code.activestate.com/recipes/499299/

    Useful for retreiving a canonical value equivalent to "item".  For example, a
    caching or interning application may require fetching a single representative
    instance from many possible equivalent instances).

    >>> get_equivalent(set([1, 2, 3]), 2.0)             # 2.0 is equivalent to 2
    2
    >>> get_equivalent([1, 2, 3], 4, default=0)
    0

    NOTE: Use this function with cautious because the direction of == can go
    in different directions, and sometimes the _CaptureEq.__eq__ is not called.
    moPepGen.aa.AminoAcidSeqRecord.__eq__ is an example of a work around.
    '''
    t = _CaptureEq(item)

    if t in container:
        return t.match
    return default


def logger(message:str) -> None:
    """ Print message to the stdout. """
    print(
        f'[ {datetime.now().strftime(format="%Y-%m-%d %H:%M:%S")} ] {message}',
        flush=True
    )

def get_logger() -> logging.Logger:
    """ get logger """
    return logging.getLogger(constant.PROG_NAME)

def all_equal(iterable:Iterable) -> bool:
    """ Check if all elements are equal """
    it = itertools.groupby(iterable)
    return next(it, True) and not next(it, False)

def check_sha512(handle:IO):
    """ check sha512sum. Handle must be readable """
    sum_val = hashlib.sha512()
    # pylint: disable=W0640
    for byte_block in iter(lambda: handle.read(4096), b""):
        sum_val.update(byte_block)
    return sum_val.hexdigest()



""" Module for ENAEdge class """
from __future__ import annotations
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .TVGNode import TVGNode

class TVGEdge():
    """ Defines the edges in the TranscriptVariantGraph

    Attributes:
        in_node (Node): The inbond node.
        out_node (Node): The outbond node.
        type (str): The edge type. Must be either of orf_start, orf_end,
            variant_start, variant_end, cleave, or reference
    """
    def __init__(self, in_node:TVGNode, out_node:TVGNode,
            _type:str):
        """ Constructor for Edge

        Args:
            in_node (Node): The inbond node.
            out_node (Node): The outbond node.
            type (str): The edge type. Must be either of orf_start, orf_end,
                mutation_start, mutation_end, cleave, or reference
        """
        self.in_node = in_node
        self.out_node = out_node
        edge_types = ['variant_start', 'variant_end', 'reference']
        if _type not in edge_types:
            raise ValueError(f'type {_type} not from {edge_types}')
        self.type = _type



""" This module defined classes in order to group certain parameters together. """
from __future__ import annotations
from typing import TYPE_CHECKING, Set

from moPepGen import aa


if TYPE_CHECKING:
    from moPepGen import dna, gtf


class CleavageParams():
    """ Cleavage related parameters.

    ## Attributes:
        - enzyme (str): Enzyme name.
        - exception (str): Enzymatic cleavage exception.
        - miscleavage (int): Number of cleavages to allow per non-canonical peptide.
        - min_mw (int): The minimal molecular weight of the non-canonical peptides.
        - min_length (int): The minimal length of non-canonical peptides, inclusive.
        - max_length (int): The maximum length of non-canonical peptides, inclusive.
        - max_variants_per_node (int): Maximal number of variants per node. This
            can be useful when there are local regions that are heavily mutated.
            When creating the cleavage graph, nodes containing variants larger
            than this value are skipped. Setting to -1 will avoid checking for
            this.
        - additional_variants_per_misc (int): Additional variants allowed for
            every miscleavage. This argument is used together with
            max_variants_per_node to handle hypermutated regions. Setting to -1
            will avoid checking for this.
        - min_nodes_to_collapse (int): When making the cleavage graph, the minimal
            number of nodes to trigger pop collapse.
        - naa_to_collapse (int): The number of bases used for pop collapse.
    """
    def __init__(self, enzyme:str=None, exception:str=None, miscleavage:int=2,
            min_mw:int=500, min_length:int=7, max_length:int=25,
            max_variants_per_node:int=7, additional_variants_per_misc:int=2,
            min_nodes_to_collapse:int=30, naa_to_collapse:int=5):
        """ constructor """
        self.enzyme = enzyme
        self.exception = exception
        self.miscleavage = miscleavage
        self.min_mw = min_mw
        self.min_length = min_length
        self.max_length = max_length
        self.max_variants_per_node = max_variants_per_node
        self.additional_variants_per_misc = additional_variants_per_misc
        self.min_nodes_to_collapse = min_nodes_to_collapse
        self.naa_to_collapse = naa_to_collapse
        if self.exception == 'auto':
            if enzyme == 'trypsin':
                self.exception = 'trypsin_exception'
            else:
                self.exception = None

    def jsonfy(self, graph_params:bool=False):
        """ jsonfy """
        data = {
            'enzyme': self.enzyme,
            'exception': self.exception,
            'miscleavage': self.miscleavage,
            'min_mw': self.min_mw,
            'min_length': self.min_length,
            'max_length': self.max_length
        }
        if graph_params:
            data.update({
                'max_variants_per_node': self.max_variants_per_node,
                'additional_variants_per_misc': self.additional_variants_per_misc,
                'min_nodes_to_collapse': self.min_nodes_to_collapse,
                'naa_to_collapse': self.naa_to_collapse
            })
        return data

class ReferenceData():
    """ Reference related parameters

    ## Attributes
        - genome (dna.DNASeqDict)
        - anno (gtf.GeneAnnotationModel)
        - canonical_peptides (Set[str])
        - proteome (aa.AminoAcidSeqDict)
    """
    def __init__(self, genome:dna.DNASeqDict, anno:gtf.GenomicAnnotation,
            canonical_peptides:Set[str], proteome:aa.AminoAcidSeqDict=None):
        """ constructor """
        self.genome = genome
        self.anno = anno
        self.canonical_peptides = canonical_peptides
        self.proteome = proteome  or aa.AminoAcidSeqDict()
